/*  _ __ ___ _
 * | |\ /  /| |  $Id: CdlFile.C,v 1.7 1997/08/11 02:27:02 dpwe Exp $
 * | | /  / | |  Copyright (C) 1995 IXI Limited.
 * |_|/__/_\|_|  IXI Limited, Cambridge, England.
 *
 * Component   : CdlFile.C
 *
 * Author      : Dean Sheehan (deans@x.co.uk)
 *
 * Description : Contains the implementation for the CdlFile class that 
 *               models the CDL file read.
 *
 * License     :
			Object Tcl License & Copyright
			------------------------------

IXI Object Tcl software, both binary and source (hereafter, Software) is copyrighted by IXI Limited (IXI), and ownership remains with IXI. 

IXI grants you (herafter, Licensee) a license to use the Software for academic, research and internal business purposes only, without a fee. Licensee may distribute the binary and source code (if required) to third parties provided that the copyright notice and this statement appears on all copies and that no charge is associated with such copies. 

Licensee may make derivative works. However, if Licensee distributes any derivative work based on or derived from the Software, then Licensee will (1) notify IXI regarding its distribution of the derivative work, and (2) clearly notify users that such derivative work is a modified version and not the original IXI Object Tcl distributed by IXI. IXI strongly recommends that Licensee provide IXI the right to incorporate such modifications into future releases of the Software under these license terms. 

Any Licensee wishing to make commercial use of the Software should contact IXI, to negotiate an appropriate license for such commercial use. Commercial use includes (1) integration of all or part of the source code into a product for sale or license by or on behalf of Licensee to third parties, or (2) distribution of the binary code or source code to third parties that need it to utilize a commercial product sold or licensed by or on behalf of Licensee. 

IXI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY. IXI SHALL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER SUFFERED BY THE USERS OF THIS SOFTWARE. 

Copyright (C) 1995, IXI Limited 

By using or copying this Software, Licensee agrees to abide by the copyright law and all other applicable laws of England and the U.S., including, but not limited to, export control laws, and the terms of this license. IXI shall have the right to terminate this license immediately by written notice upon Licensee's breach of, or non-compliance with, any of its terms. Licensee may be held legally responsible for any copyright infringement that is caused or encouraged by Licensee's failure to abide by the terms of this license. 

Comments and questions are welcome and can be sent to
otcl@x.co.uk 

For more information on copyright and licensing issues, contact: 
Legal Department, IXI Limited, Vision Park, Cambridge CB4 4ZR,
ENGLAND. 

 *
 */

// System Includes
#include <string.h>
#include <iostream.h>

// Tcl Includes
#include <tcl.h>
#include <tclInt.h>

// Local Includes
#include "CdlFile.H"
#include "CdlItem.H"
#include "CdlMethod.H"
#include "CdlSlot.H"
#include "CdlArg.H"
#include "CdlRtn.H"

CdlFile::CdlFile (char *n)
{
   name = strdup(n);
   noOfItems = 0;
   interp = Tcl_CreateInterp();
   methodUnderConstruction = NULL;
   classUnderConstruction = NULL;
}

CdlFile::~CdlFile ()
{
   for (int c = 0; c < noOfItems; c++)
   {
      delete item[c];
   }
   Tcl_DeleteInterp(interp);
}

int CdlFile::parse (void)
{
   // Thought it would be a good idea once but people complained
   // rmvStdTclCmds();

   addItemCmds();

   int res = Tcl_EvalFile(interp,name);
   if (res != TCL_OK)
   {
      cerr << "Error in input file: ";
      cerr << Tcl_GetVar(interp,"errorInfo",0) << endl;
      return 0;
   }

   return 1;
}

int CdlFile::genHeader (char *n)
{
   ofstream of(n);
   if (!of)
   {
      return 0;
   }

   genHeaderStart (of);

   for (int i = 0; i < noOfItems; i++)
   {
      item[i]->genHeader(of);
      of << endl;
   }

   genHeaderEnd (of);

   return 1;
}

int CdlFile::genSource (char *n)
{
   ofstream of(n);
   if (!of)
   {
      return 0;
   }

   genSourceStart (of);
 
   for (int i = 0; i < noOfItems; i++)
   {
      item[i]->genSource(of);
      of << endl;
   }

   genSourceEnd (of);

   return 1;
}

void CdlFile::rmvStdTclCmds (void)
{
   // This is a really nasty way of doing it, but...
   Interp *i = (Interp*)interp;
   Tcl_InitHashTable(&i->commandTable,TCL_STRING_KEYS);
}

void CdlFile::genHeaderStart (ofstream &of)
{
    /* find the tail part of the name (for the DEF symbol) */
   char *nametail = strrchr(name, '/');
   if (nametail) {
       // found a "/"; step to next chr
       ++nametail;
   } else {
       // no "/", so use the whole thing
       nametail = name;
   }
   char *tmpName = strdup(nametail);
   for (char *c = tmpName; *c !=NULL; c++)
   {
      if (*c == '.')
      {
         *c = '_';
      }
   }

   of << "#ifndef " << tmpName << endl;
   of << "#define " << tmpName << endl;

   of << endl;

   of << "#include <iostream.h>" << endl;
   of << "#include <stdlib.h>" << endl;
   of << "#include <string.h>" << endl;
   of << "#include <Otcl.H>" << endl;
   of << "#include <OtclObject.H>" << endl;
   of << "#include <OtclPart.H>" << endl;
   of << endl;

}

void CdlFile::genHeaderEnd (ofstream &of)
{
   of << endl;
   of << "#endif" << endl;
}

void CdlFile::genSourceStart (ofstream &of)
{
}

void CdlFile::genSourceEnd (ofstream &)
{
}

void CdlFile::addItemCmds (void)
{
   Tcl_CreateCommand(interp,"class",CdlFile::classCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"pass",CdlFile::passCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"global",CdlFile::globalCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
}

void CdlFile::rmvItemCmds (void)
{
   Tcl_DeleteCommand(interp,"class");
   Tcl_DeleteCommand(interp,"pass");
   Tcl_DeleteCommand(interp,"global");
}

void CdlFile::addMethodCmds (void)
{
   Tcl_CreateCommand(interp,"method",CdlFile::methodCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"slot",CdlFile::slotCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"classMethod",CdlFile::classMethodCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"classSlot",CdlFile::classSlotCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"constructor",CdlFile::constructorCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);

}

void CdlFile::rmvMethodCmds (void)
{
   Tcl_DeleteCommand(interp,"method");
   Tcl_DeleteCommand(interp,"slot");
   Tcl_DeleteCommand(interp,"classMethod");
   Tcl_DeleteCommand(interp,"classSlot");
   Tcl_DeleteCommand(interp,"constructor");
}

void CdlFile::addArgCmds (void)
{
   Tcl_CreateCommand(interp,"void",CdlFile::argCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   CdlArg *current = CdlArg::exemplarHead;
   while (current != NULL)
   {
      Tcl_CreateCommand(interp,current->giveCommand(),CdlFile::argCmd,
                        (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
      current = current->nextExemplar;
 
   }
}

void CdlFile::rmvArgCmds (void)
{
   Tcl_DeleteCommand(interp,"void");
   CdlArg *current = CdlArg::exemplarHead;
   while (current != NULL)
   {
      Tcl_DeleteCommand(interp,current->giveCommand());
      current = current->nextExemplar;
   }
}

void CdlFile::addRtnCmds (void)
{
   Tcl_CreateCommand(interp,"void",CdlFile::rtnCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
   CdlRtn *current = CdlRtn::exemplarHead;
   while (current != NULL)
   {
      Tcl_CreateCommand(interp,current->giveCommand(),CdlFile::rtnCmd,
                        (ClientData)this,(Tcl_CmdDeleteProc*)NULL);
      current = current->nextExemplar;
   }
}

void CdlFile::rmvRtnCmds (void)
{
   Tcl_DeleteCommand(interp,"void");
   CdlRtn *current = CdlRtn::exemplarHead;
   while (current != NULL)
   {
      Tcl_DeleteCommand(interp,current->giveCommand());
      current = current->nextExemplar;
   }
}

int CdlFile::classCmd (ClientData cd, Tcl_Interp *, int argc, char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->newClass(argc,argv);
}

int CdlFile::globalCmd (ClientData cd, Tcl_Interp *, int argc, char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->global(argc,argv);
}

int CdlFile::passCmd (ClientData cd, Tcl_Interp *, int argc, char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->pass(argc,argv);
}

int CdlFile::constructorCmd (ClientData cd, Tcl_Interp *, int argc,
                               char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->constructor(argc,argv);
}

int CdlFile::methodCmd (ClientData cd, Tcl_Interp *, int argc, char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->method(argc,argv);
}

int CdlFile::slotCmd (ClientData cd, Tcl_Interp *, int argc, char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->slot(argc,argv);
}

int CdlFile::classMethodCmd (ClientData cd, Tcl_Interp *, int argc,
                               char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->classMethod(argc,argv);
}

int CdlFile::classSlotCmd (ClientData cd, Tcl_Interp *, int argc,
                               char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->classSlot(argc,argv);
}

int CdlFile::argCmd (ClientData cd, Tcl_Interp *, int argc,
                               char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->arg(argc,argv);
}

int CdlFile::rtnCmd (ClientData cd, Tcl_Interp *, int argc,
                               char *argv[])
{
   CdlFile *cdlF = (CdlFile*)cd;
   return cdlF->rtn(argc,argv);
}

int CdlFile::newClass (int argc, char *argv[])
{
   char **superclasses = NULL;
   int noOfSuperclasses = 0;
   char *body = NULL;
   int castable = 0;		// dpwe
   int i = 0;
   int n;
   int justfwd = 0;
   int res = TCL_OK;

   if (argc == 2) {
       /* just a fwd decln */
       justfwd = 1;
   } else if (argc < 3 || argc >6) {
      Tcl_SetResult(interp,"wrong # args",TCL_STATIC);
      return TCL_ERROR;
   }

//   fprintf(stderr, "newClass for %s (args %d, noOfItems %d)\n", 
//	   argv[1], argc, noOfItems);

   i = 2;
   while(i < argc-1)  {
       n = strlen(argv[i]);
       if( strncmp(argv[i], "-castable", n)==0 )  {
	   castable = 1;
	   ++i;
       } else if(strncmp(argv[i], "-isA", n)==0)  {
	   if(i+1 > argc-2)  {
	       Tcl_SetResult(interp,"too few arguments to \"class -isA\" <parent-list> <body>",TCL_STATIC);
	       return TCL_ERROR;
	   }
	   Tcl_SplitList(interp,argv[i+1],&noOfSuperclasses,&superclasses);
	   i += 2;
       } else {
	   Tcl_SetResult(interp,"unrecognized argument to class not \"-castable\" or \"-isA\"",TCL_STATIC);
	   return TCL_ERROR;
       }
   }
   body = argv[i];

   char **namelist = NULL;
   int noNames;
   Tcl_SplitList(interp, argv[1], &noNames, &namelist);

   CdlClass *gclass = CdlClass::findClass(namelist[0]);
   if(gclass == NULL) {
       if(noOfItems == MAX_ITEMS) {
	   char n[32];
	   sprintf(n, "%d", noOfItems);
	   Tcl_AppendResult(interp, "at class ", namelist[0], 
			    ": too many classes in file (", n, ")!", 
			    NULL);
	   return TCL_ERROR;
       }
       gclass = new CdlClass(namelist[0]);
       item[noOfItems] = gclass;
       noOfItems++;
   } 
   if(noNames > 1) {
       gclass->setCname(namelist[1]);
   }
   classUnderConstruction = gclass;

   if(!justfwd) {
       if(!gclass->ReInit(noOfSuperclasses, superclasses, castable)) {
	   Tcl_AppendResult(interp,"cannot redeclare class \"", namelist[0], 
			    "\"", NULL);
	   return TCL_ERROR;
       }

       rmvItemCmds();
       
       addMethodCmds();

       res = Tcl_Eval(interp,body);
       
       rmvMethodCmds ();

       addItemCmds(); 

       if (res == TCL_OK \
	   && classUnderConstruction->getConstructor() == NULL \
	   && classUnderConstruction->doesInherit() )   {
	   // Set up default constructor if there is not one defined but
	   // we are an inheriting class, so we need one.
	   char n[50];
	   sprintf(n,"%s%s", classUnderConstruction->giveName(),
		   OTCL_PART_SUFFIX);
	   classUnderConstruction->setConstructor(
	                                  new CdlConstructor(n, castable));
       }
   }

   return res;
}

int CdlFile::global (int argc, char *argv[])
{   /* hacked version of newClass for global commands */
   char **superclasses = NULL;
   int noOfSuperclasses = 0;
   char *body = NULL;
   int castable = 0;		// dpwe
   int i = 0;
   int n;

   if (argc != 2)  {
      Tcl_SetResult(interp,"wrong # args",TCL_STATIC);
      return TCL_ERROR;
   }

   castable = 1;
   body = argv[1];

   if(noOfItems == MAX_ITEMS) {
       char n[32];
       sprintf(n, "%d", noOfItems);
       Tcl_AppendResult(interp, "at global: too many classes in file (", n, ")!", 
			NULL);
       return TCL_ERROR;
   }
   CdlClass *gclass = new CdlClass("G");	// the name to access
   gclass->ReInit(0,NULL,castable);
   // make all the C references harmless
   gclass->setCname("");	// "void" breaks horribly
   item[noOfItems] = gclass;
   noOfItems++;
   classUnderConstruction = gclass;

   rmvItemCmds();

   addMethodCmds();

   int res = Tcl_Eval(interp,body);

   rmvMethodCmds ();

   addItemCmds(); 

   if (res == TCL_OK \
       && classUnderConstruction->getConstructor() == NULL \
       && classUnderConstruction->doesInherit() )  {
      // Set up default constructor if there is not one defined but
      // we are an inheriting class, so we need one.
      char n[50];
      sprintf(n,"%s%s", classUnderConstruction->giveName(),
              OTCL_PART_SUFFIX);
      classUnderConstruction->setConstructor(new CdlConstructor(n, castable));
   }

   return res;
}

int CdlFile::pass (int argc, char *argv[])
{
   if (argc < 2 || argc > 3)
   {
      // error
   }
   
   if(noOfItems == MAX_ITEMS) {
       char n[32];
       sprintf(n, "%d", noOfItems);
       Tcl_AppendResult(interp, "at pass: too many classes in file (", 
			n, ")!", NULL);
       return TCL_ERROR;
   }
   if (argc == 2) {
      item[noOfItems] = new CdlPass(argv[1]);
   } else  {
      item[noOfItems] = new CdlPass(argv[1],argv[2]);
   }
   noOfItems++;
   return TCL_OK;
}

int CdlFile::constructor (int argc, char *argv[])
{
   if (argc != 2)
   {
      Tcl_SetResult(interp,"wrong # args",TCL_STATIC);
      return TCL_ERROR;
   }

   CdlClass *c = classUnderConstruction;

   char cname[50];
   sprintf(cname,"%s%s",c->giveName(),OTCL_PART_SUFFIX);
   methodUnderConstruction = new CdlConstructor(cname, !c->doesInherit());

   rmvMethodCmds();
   addArgCmds();

   int res = Tcl_Eval(interp,argv[1]);

   rmvArgCmds();
   addMethodCmds();

   if (res == TCL_OK)
   {
      c->setConstructor((CdlConstructor*)methodUnderConstruction);
      methodUnderConstruction = NULL;
      res = c->constructorOK();
   }

   return res;
}

int CdlFile::method (int argc, char *argv[])
{
    char *argl;
    char *retl;

   if (argc < 4 || argc > 6) {
      Tcl_SetResult(interp,"wrong # args: \"method <name> [-static|-dynamic|-dummy] <args> <return>\"",TCL_STATIC);
      return TCL_ERROR;
   }

//   fprintf(stderr, "declaring method %s::%s\n", 
// 	    classUnderConstruction->giveName(), argv[1]);

   CdlInstanceMethod *m = new CdlInstanceMethod(argv[1]);
   methodUnderConstruction = m;

   if (strcmp(argv[2],STATIC_FLAG) == 0) {
      if (argc != 5)  {
	  Tcl_SetResult(interp,"wrong # args: \"method <name> -static <args> <return>\"",TCL_STATIC);
	  return TCL_ERROR;
      }
      m->setBinding(CdlMethod::STATIC);
      argl = argv[3];
      retl = argv[4];
   } else if (strcmp(argv[2],DYNAMIC_FLAG) == 0) {
      if (argc != 5)  {
	  Tcl_SetResult(interp,"wrong # args: \"method <name> -dynamic <args> <return>\"",TCL_STATIC);
	  return TCL_ERROR;
      }
      m->setBinding(CdlMethod::DYNAMIC);
      argl = argv[3];
      retl = argv[4];
   } else if (strcmp(argv[2],"-dummy") == 0) {
      if (argc != 4)  {
	  Tcl_SetResult(interp,"wrong # args: \"method <name> -dummy <script>\"",TCL_STATIC);
	  return TCL_ERROR;
      }
      m->setBinding(CdlMethod::DUMMYTCL);
      argl = "";
      retl = "";
      m->setScript(argv[3]);
   } else if (strcmp(argv[2],"-dummyCC") == 0) {
      if (argc != 6)  {
	  Tcl_SetResult(interp,"wrong # args: \"method <name> -dummyCC <args> <return> <c-code>\"",TCL_STATIC);
	  return TCL_ERROR;
      }
      m->setBinding(CdlMethod::DUMMYCPP);
      argl = argv[3];
      retl = argv[4];
      m->setScript(argv[5]);
   } else {
//      Tcl_SetResult(interp,"unknown 3rd argument: must be \"static\" or \"dynamic\"",TCL_STATIC);
//      return TCL_ERROR;
       if (argc != 4)  {
	   Tcl_SetResult(interp,"wrong # args: \"method <name> <args> <return>\"",TCL_STATIC);
	   return TCL_ERROR;
       }
       m->setBinding(CdlMethod::STATIC);	// default to static binding
       argl = argv[2];		// point to right remainder components
       retl = argv[3];
   }

   rmvMethodCmds();
   addArgCmds();

   int res = Tcl_Eval(interp,argl);

   rmvArgCmds();

   if (res == TCL_OK)
   {
      addRtnCmds();

      res = Tcl_Eval(interp,retl);

      rmvRtnCmds();
   }

   addMethodCmds();

   if (res == TCL_OK)
   {
      classUnderConstruction->addInstanceMethod(m);
      methodUnderConstruction = NULL;
   }

   if (m->getRtn() == NULL && !m->returnIsVoid())
   {
      m->setRtn(new CdlIntRtn());
   }

   return res;
}

int CdlFile::slot (int argc, char *argv[])
{
    int res = TCL_OK;
    int isfake = 0;

    if (argc < 3 || argc > 6)  {
	Tcl_SetResult(interp,"wrong # args: \"slot <name> ?-dummy? <type> ?<set script> ?<get script>??\"",
		      TCL_STATIC);
	return TCL_ERROR;
    }

    int n, i = 2;
    while(i < argc && argv[i][0] == '-') {
	n = strlen(argv[i]);
	if( strncmp(argv[i], "-dummy", n)==0 )  {
	    isfake = 1;
	} else {
	    Tcl_AppendResult(interp, "unrecognized slot option \"", argv[i], "\" not \"-dummy\"", NULL);
	    return TCL_ERROR;
	}
	++i;
    }
    CdlSlot *m;
    if (isfake)
	m = new CdlFakeSlot(argv[1]);
    else
	m = new CdlSlot(argv[1]);
    methodUnderConstruction = m;

    rmvMethodCmds();
    addRtnCmds();

    res = Tcl_Eval(interp,argv[i]);

    rmvRtnCmds();
    if(res == TCL_OK)  {
	addArgCmds();

	res = Tcl_Eval(interp,argv[i]);

	rmvArgCmds();
    }
    addMethodCmds();

    // Pick up setscript & getscript, if any
    ++i;
    if(i<argc) {
	m->setSetScript(argv[i]);
	++i;
	if(i<argc) {
	    m->setGetScript(argv[i]);
	    ++i;
	    if(i<argc) {
		delete m;
		methodUnderConstruction = NULL;
		Tcl_SetResult(interp,"wrong # args: \"slot <name> ?-dummy? <type> ?<set script> ?<get script>??\"",
			      TCL_STATIC);
		return TCL_ERROR;
	    }
	}
    }

    if (res == TCL_OK)   {
	if(isfake) {
	    // ensure that -default specified
	    if (!((CdlFakeSlot*)m)->isOK(interp)) {
		return TCL_ERROR;
	    }
	    classUnderConstruction->addFakeSlot((CdlFakeSlot*)m);
	} else {
	    classUnderConstruction->addSlot(m);
	}
	methodUnderConstruction = NULL;
    }

    // Because we inherited slots from methods, this is the default code 
    // to set the return value to int.  But I think it can't happen
    if (m->getRtn() == NULL && !m->returnIsVoid())   {
	m->setRtn(new CdlIntRtn());
    }

    return res;
}

int CdlFile::classMethod (int argc, char *argv[])
{
    char *argl;
    char *retl;

    if (argc < 4 || argc > 6) {
	Tcl_SetResult(interp,"wrong # args",TCL_STATIC);
	return TCL_ERROR;
    }

    CdlClassMethod *m = new CdlClassMethod(argv[1]);
    methodUnderConstruction = m;
   
    if (strcmp(argv[2],"-dummy") == 0) {
	if (argc != 4)  {
	    Tcl_SetResult(interp,"wrong # args: \"classMethod <name> -dummy <script>\"",TCL_STATIC);
	    return TCL_ERROR;
	}
	m->setBinding(CdlMethod::DUMMYTCL);
	argl = "";
	retl = "";
	m->setScript(argv[3]);
    } else if (strcmp(argv[2],"-dummyCC") == 0) {
	if (argc != 6)  {
	    Tcl_SetResult(interp,"wrong # args: \"classMethod <name> -dummyCC <args> <return> <c-code>\"",TCL_STATIC);
	    return TCL_ERROR;
	}
	m->setBinding(CdlMethod::DUMMYCPP);
	argl = argv[3];
	retl = argv[4];
	m->setScript(argv[5]);
    } else {
	if (argc != 4)  {
	    Tcl_SetResult(interp,"wrong # args: \"classMethod <name> <args> <return>\"",TCL_STATIC);
	    return TCL_ERROR;
	}
	m->setBinding(CdlMethod::STATIC);	// default to static binding
	argl = argv[2];		// point to right remainder components
	retl = argv[3];
    }

    rmvMethodCmds();
    addArgCmds();

    int res = Tcl_Eval(interp,argl);

    rmvArgCmds();

    if (res == TCL_OK) {
	addRtnCmds();
	res = Tcl_Eval(interp,retl);
	rmvRtnCmds();
    }
    
    if (res == TCL_OK) {
	classUnderConstruction->addClassMethod(m);
	methodUnderConstruction = NULL;
    }
    
    if (m->getRtn() == NULL && !m->returnIsVoid()) {
	m->setRtn(new CdlIntRtn());
    }

    addMethodCmds();

    return res;
}

int CdlFile::classSlot (int argc, char *argv[])
{
    int res = TCL_OK;

    if (argc != 3)  {
	Tcl_SetResult(interp,"wrong # args: \"classSlot <name> <type>\"",
		      TCL_STATIC);
	return TCL_ERROR;
    }

    CdlClassSlot *m = new CdlClassSlot(argv[1]);
    methodUnderConstruction = m;

    rmvMethodCmds();
    addRtnCmds();

    res = Tcl_Eval(interp,argv[2]);

    rmvRtnCmds();
    if(res == TCL_OK)  {
	addArgCmds();

	res = Tcl_Eval(interp,argv[2]);

	rmvArgCmds();
    }
    addMethodCmds();

    if (res == TCL_OK)   {
	classUnderConstruction->addClassSlot(m);
	methodUnderConstruction = NULL;
    }

    // Because we inherited slots from methods, this is the default code 
    // to set the return value to int.  But I think it can't happen
    if (m->getRtn() == NULL && !m->returnIsVoid())   {
	m->setRtn(new CdlIntRtn());
    }

    return res;
}

int CdlFile::arg (int argc, char *argv[])
{
   if (methodUnderConstruction->argsAreVoid())
   {
      return TCL_ERROR;
   }

   if (strcmp(argv[0],"void") == 0)
   {
      methodUnderConstruction->setVoidArgs();
      return TCL_OK;
   }

   CdlArg *current = CdlArg::exemplarHead;
   while (current != NULL)
   {
      if (strcmp(current->giveCommand(),argv[0]) == 0)
      {
         current = current->instantiate();
         if (!current->setArgs(argc-1,&argv[1]))
         {
            return TCL_ERROR;
         }

         if(!methodUnderConstruction->addArg(current)) {
	     return TCL_ERROR;
	 }
         return TCL_OK;
      }
      current = current->nextExemplar;
   }

   return TCL_OK;
}

int CdlFile::rtn (int argc, char *argv[])
{
   if (methodUnderConstruction->getRtn() != NULL)
   {
      return TCL_ERROR;
   }
   if (methodUnderConstruction->returnIsVoid())
   {
      return TCL_ERROR;
   }

   if (strcmp(argv[0],"void") == 0)
   {
      methodUnderConstruction->setVoidReturn();
      return TCL_OK;
   }

   CdlRtn *current = CdlRtn::exemplarHead;
   while (current != NULL)
   {
      if (strcmp(current->giveCommand(),argv[0]) == 0)
      {
         current = current->instantiate();
         if (!current->setArgs(argc-1,&argv[1]))
         {
            return TCL_ERROR;
         }
         
         methodUnderConstruction->setRtn(current);
         return TCL_OK;
      }
      current = current->nextExemplar;
   }

   return TCL_ERROR;
}
