/*  _ __ ___ _
 * | |\ /  /| |  $Id: CdlItem.C,v 1.7 1997/08/11 02:27:02 dpwe Exp $
 * | | /  / | |  Copyright (C) 1995 IXI Limited.
 * |_|/__/_\|_|  IXI Limited, Cambridge, England.
 *
 * Component   : CdlItem.C
 *
 * Author      : Dean Sheehan (deans@x.co.uk)
 *
 * Description : Contains the implementation for CdlClass and CdlPass, the two
 *               top level entities within a CDL file.
 *
 * License     :
			Object Tcl License & Copyright
			------------------------------

IXI Object Tcl software, both binary and source (hereafter, Software) is copyrighted by IXI Limited (IXI), and ownership remains with IXI. 

IXI grants you (herafter, Licensee) a license to use the Software for academic, research and internal business purposes only, without a fee. Licensee may distribute the binary and source code (if required) to third parties provided that the copyright notice and this statement appears on all copies and that no charge is associated with such copies. 

Licensee may make derivative works. However, if Licensee distributes any derivative work based on or derived from the Software, then Licensee will (1) notify IXI regarding its distribution of the derivative work, and (2) clearly notify users that such derivative work is a modified version and not the original IXI Object Tcl distributed by IXI. IXI strongly recommends that Licensee provide IXI the right to incorporate such modifications into future releases of the Software under these license terms. 

Any Licensee wishing to make commercial use of the Software should contact IXI, to negotiate an appropriate license for such commercial use. Commercial use includes (1) integration of all or part of the source code into a product for sale or license by or on behalf of Licensee to third parties, or (2) distribution of the binary code or source code to third parties that need it to utilize a commercial product sold or licensed by or on behalf of Licensee. 

IXI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY. IXI SHALL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER SUFFERED BY THE USERS OF THIS SOFTWARE. 

Copyright (C) 1995, IXI Limited 

By using or copying this Software, Licensee agrees to abide by the copyright law and all other applicable laws of England and the U.S., including, but not limited to, export control laws, and the terms of this license. IXI shall have the right to terminate this license immediately by written notice upon Licensee's breach of, or non-compliance with, any of its terms. Licensee may be held legally responsible for any copyright infringement that is caused or encouraged by Licensee's failure to abide by the terms of this license. 

Comments and questions are welcome and can be sent to
otcl@x.co.uk 

For more information on copyright and licensing issues, contact: 
Legal Department, IXI Limited, Vision Park, Cambridge CB4 4ZR,
ENGLAND. 

 *
 */

// System Includes
#include <string.h>
#include <stdlib.h>

// Tcl Includes

// Local Includes
#include "CdlItem.H"
#include "CdlMethod.H"
#include "CdlSlot.H"
#include "CdlFile.H"

CdlIndent::CdlIndent (int n)
{
   spaces = TAB * n;
}

CdlIndent &CdlIndent::operator ++ (int)
{
   spaces += TAB;
   return *this;
}

CdlIndent &CdlIndent::operator -- (int)
{
   spaces -= TAB;
   return *this;
}

ofstream &operator << (ofstream &o, CdlIndent &c)
{
   int i;
   for (i = 0; i < c.spaces; i++)
   {
      o << " ";
   }
   return o;
}

CdlItem::CdlItem ()
{
}

CdlItem::~CdlItem ()
{
}

CdlPass::CdlPass (char *l)
{
   pos = HEADER_AND_SOURCE;

   if (*l == '\n')
   {
      l++;
   }
   line = strdup(l);
   if (line[strlen(line) - 1] == '\n')
   {
      line[strlen(line) - 1] = NULL;
   }
}

CdlPass::CdlPass (char *p, char *l)
{
   pos = HEADER_AND_SOURCE;

   if (strcmp(p,"-h") == 0)
   {
      pos = HEADER;
   }
   else if (strcmp(p,"-s") == 0)
   {
      pos = SOURCE;
   }

   if (*l == '\n')
   {
      l++;
   }
   line = strdup(l);
   if (line[strlen(line) - 1] == '\n')
   {
      line[strlen(line) - 1] = NULL;
   }
}

CdlPass::~CdlPass ()
{
   free(line);
}

int CdlPass::genHeader (ofstream &o)
{
   if (pos == HEADER || pos == HEADER_AND_SOURCE)
   {
      o << line << endl;
   }
   return 0;
}

int CdlPass::genSource (ofstream &o)
{
   if (pos == SOURCE || pos == HEADER_AND_SOURCE)
   {
      o << line << endl;
   }
   return 0;
}

classRecdNode *CdlClass::classlist = NULL;

CdlClass::CdlClass (char *n)
{
   name = strdup(n);
   cname = name;
   noOfInstanceMethods = 0;
   noOfSlots = 0;
   noOfFakeSlots = 0;
   noOfClassMethods = 0;
   noOfClassSlots = 0;
   constructor = NULL;
   noOfSuperclasses = 0;
   superclasses = NULL;
   castable = 0;
   /* add it into our table */
   addClass(name, this);
   /* but don't you believe it */
   dummy = 1;
}

int CdlClass::ReInit (int noOfS, char *s[], int castbl)
{	/* non-forward declaration reinit */
    /* If we've used this one at all, object */
//    fprintf(stderr, "ReInit of %s: nMethods = %d\n", 
//	    name, noOfInstanceMethods);
    if (!dummy) {
	return 0;
    } else {
	noOfSuperclasses = noOfS;
	superclasses = s;
	castable = castbl;
	dummy = 0;
	return 1;
    }
}

CdlClass::~CdlClass ()
{
    /* remove from static table */
    rmvClass(name);
    
    if(cname != NULL && cname != name)	free(cname);

   free(name);
   if (constructor != NULL)
   {
      delete constructor;
   }
   int m;
   for (m = 0; m < noOfInstanceMethods; m++)
   {
      delete instanceMethod[m];
   }
   for (m = 0; m < noOfSlots; m++)
   {
      delete slot[m];
   }
   for (m = 0; m < noOfFakeSlots; m++)
   {
      delete fakeSlot[m];
   }
   for (m = 0; m < noOfClassMethods; m++)
   {
      delete classMethod[m];
   }
   for (m = 0; m < noOfClassSlots; m++)
   {
      delete classSlot[m];
   }
   if (noOfSuperclasses != 0)
   {
      free(superclasses);
   }
}

char *CdlClass::giveName (void)
{
   return name;
}

void CdlClass::addInstanceMethod (CdlInstanceMethod *m)
{
    dummy = 0;
   instanceMethod[noOfInstanceMethods] = m;
   noOfInstanceMethods++;
}

void CdlClass::addSlot (CdlSlot *m)
{
    dummy = 0;
   slot[noOfSlots] = m;
   noOfSlots++;
}

void CdlClass::addFakeSlot (CdlFakeSlot *m)
{
    dummy = 0;
   fakeSlot[noOfFakeSlots] = m;
   noOfFakeSlots++;
}

void CdlClass::addClassMethod (CdlClassMethod *m)
{
    dummy = 0;
   classMethod[noOfClassMethods] = m;
   noOfClassMethods++;
}

void CdlClass::addClassSlot (CdlClassSlot *m)
{
    dummy = 0;
   classSlot[noOfClassSlots] = m;
   noOfClassSlots++;
}

void CdlClass::setConstructor (CdlConstructor *c)
{
    dummy = 0;
   constructor = c;

   // Check to see if constructor is just copy constructor
//   if(c->isCopyConstructor(name))  {
//       castable = 1;
//   }
   // now a flag to CdlClass constructor
}

int CdlClass::constructorOK() 
{    // check that constructor conforms to constraints of class type (...)
    int rslt = TCL_OK;
   // At the moment, you can only use -from args in a constructor 
   // if the class is castable (an annoying limitation in the way 
   // I implemented it;  I want to be able to construct the OTCL_PART
   // companion object *before* I create the actual C++ object, and 
   // I can't do that if one is subclassed from the other (the !castable 
   // case).  
   // Check here & flag error
   if(!castable) {
       if (constructor->hasFromArgs()) {
	   fprintf(stderr, "ERROR: To use -from in a constructor, the class must be -castable (for now)\n");
	   rslt = TCL_ERROR;
       }
   }
   return rslt;
}

CdlConstructor *CdlClass::getConstructor (void)
{
   return constructor;
}

int CdlClass::genHeader (ofstream &o)
{
    if(dummy)	return 1;

   genOtclClassDec(o);
   o << endl;
   // An empty cname is the sign of the 'global' class...
   if(strlen(cname)) {
       genOtclPartDec(o);
       o << endl;
   }
   return 1;
}

int CdlClass::genSource (ofstream &o)
{
    if(dummy)	return 1;

   genOtclClassDef(o);
   o << endl;
   // An empty cname is the sign of the 'global' class...
   if(strlen(cname)) {
       genOtclPartDef(o);
       o << endl;
   }
   return 1;
}

void CdlClass::genOtclClassDec (ofstream &o)
{
   CdlIndent tab(1);

    if(dummy)	return;

   // Shared library initilizer function
   o << "extern \"C\" int " << name << "_Init(Tcl_Interp *interp);" << endl;
   o << endl;
   // Class definition
   o << "class " << name << OTCL_CLASS_SUFFIX << " : public OtclClassCpp"<<endl;
   o << "{" << endl;
   o << "public:" << endl;
   o << tab << name << OTCL_CLASS_SUFFIX << " ();" << endl;
if(constructor)  {
   o << tab << "OtclPart *instantiatePart (Tcl_Interp *, int *, int," << endl;
   o << tab << "                           char *[], OtclObject *," << endl;
   o << tab << "                           OtclPart **partPtr);" << endl;
}
   o << tab << "static " << name << OTCL_CLASS_SUFFIX << " instance;" << endl;
   o << tab << "static " << name << OTCL_CLASS_SUFFIX << " *pinstance;" << endl;
   o << tab << "int classMethod (Tcl_Interp *, int, char *[]); " << endl;
   o << tab << "char *giveClassMethods (void);" << endl;	// dpwe
   // must have this fn even for global bcz defined pure virt on base class
   o << tab << "OtclPartCpp *copyInstantiate(const void *objPtr);" << endl;
   if(noOfClassSlots > 0)  {
    o << "   int cgetFn(Tcl_Interp *interp, int argc, char *argv[]);" << endl;
    o << "   int confFn(Tcl_Interp *interp, int argc, char *argv[]);" << endl;
   }
   o << "};" << endl;
}

void CdlClass::genOtclClassDef (ofstream &of)
{
   CdlIndent tab(1);

    if(dummy)	return;

   // Static Attribute Definition
   of << name << OTCL_CLASS_SUFFIX << " " << name << OTCL_CLASS_SUFFIX
      << "::instance;" << endl;
   of << endl;
   of << name << OTCL_CLASS_SUFFIX << "* " << name << OTCL_CLASS_SUFFIX
      << "::pinstance = &";
   of << name << OTCL_CLASS_SUFFIX
      << "::instance;" << endl;
   of << endl;

   // Shared library initilizer function
   of << "extern Otcl *globalOtclPtr;" << endl << endl;
   of << "int " << name << "_Init(Tcl_Interp *interp)" << endl;
   of << "{" << endl;
   of << "    static " << name << OTCL_CLASS_SUFFIX << " *pinstance = new ";
   of << name << OTCL_CLASS_SUFFIX << ";" << endl;
   of << "    globalOtclPtr->registerOtclClassCpp(pinstance);" << endl;
   of << "    " << name << OTCL_CLASS_SUFFIX
      << "::pinstance = pinstance;" << endl;
   of << "    return TCL_OK;" << endl;
   of << "}" << endl;
   of << endl;
   // Constructor Definition
   of << name << OTCL_CLASS_SUFFIX << "::" << name << OTCL_CLASS_SUFFIX
      << " () : ";
   of << "OtclClassCpp(\"" << name << "\") {}" << endl;

   of << endl;

    // copyInstantiate - method for c++-to-otcl converter to make a new otcl
   of << "OtclPartCpp *" << name << OTCL_CLASS_SUFFIX 
      << "::copyInstantiate(const void *objPtr)" << endl;
   of << "{" << endl;
   if (strlen(cname)) {
       of << "   return (OtclPartCpp *)(new " << name << OTCL_PART_SUFFIX
	  << "((const " << cname << "&)*((" << cname << " *)objPtr)));" 
	  << endl;
   } else {
       // no instantiation for global class - just a dummy
       of << "   return NULL;" << endl;
   }
   of << "}" << endl;
   of << endl;

if(constructor)  {
   // ::instantiatePart(...)
   of << "OtclPart *" << name << OTCL_CLASS_SUFFIX << "::instantiatePart (" 
      << "Tcl_Interp *interp," << endl;
   of << tab << "int *returnCode, int argc," << endl;
   of << tab << "char *argv[], OtclObject *o," << endl;
   of << tab << "OtclPart **partPtr)" << endl;
   of << "{" << endl;

   // first, scan for configuration-style args
   of << tab << "int totargc = argc;" << endl;
   of << tab << "argc = " << constructor->getMinNoOfArgs() << ";" << endl;
   of << tab << "while (argc < totargc" << endl; tab++;
   of << tab << "&& !(argv[argc][0] == '-'" << endl; tab++;
   of << tab << "&& (argv[argc][1] < '0' || argv[argc][1] > '9')))" << endl; 
   tab--;
   of << tab << "++argc;" << endl; tab--;

   // Exception if argc not enough
//   of << tab << "if (argc != " << constructor->getNoOfArgs() << ")" << endl;
   of << tab << "if (totargc < " << constructor->getMinNoOfArgs();
   of << " || argc > " << constructor->getNoOfArgs() << ")" << endl;
   of << tab << "{" << endl;
   tab++;
   of << tab << "*returnCode = TCL_ERROR;" << endl;
   of << tab << "Otcl::setTclResult(interp,ARGS_METHOD_ERR,\"constructor\","
      << "\"" << name << "\");" << endl;
   of << tab << "return NULL;" << endl;
   tab--;
   of << tab << "}" << endl;

   if(castable) {
       // Grab the 'dummy' config args first, then init, then config again
       of << tab << name << OTCL_PART_SUFFIX << "* part = new " \
	  << name << OTCL_PART_SUFFIX " (o);" << endl;
       of << tab << "if(argc < totargc) {" << endl; tab++;
       of << tab << "*returnCode = part->confFn(interp, totargc-argc, argv+argc);" << endl;
       // confFn may set error code & error msg; just carry on
       tab--; of << tab << "}" << endl;
       // Now create the new object
       constructor->preProcessArgs(of,tab);
       of << tab << "part->newObject(";
       constructor->passArgsUpwards(of);
       of << ");" << endl;
       constructor->postProcessArgs(of,tab);
   } else {
       // subclassed, not `castable' - do the old thing
       constructor->preProcessArgs(of,tab);
 
       of << tab << name << OTCL_PART_SUFFIX << "* part = new " \
	  << name << OTCL_PART_SUFFIX << " (";
       constructor->passArgs(of,"o");

       of << ");" << endl;
       
       constructor->postProcessArgs(of,tab);

   }
   of << tab << "OtclPart *res = (OtclPart *)part;" << endl;
   of << tab << "*partPtr = res;" << endl;

   // now call config for config args
   of << tab << "if(argc < totargc) {" << endl; tab++;
   of << tab << "*returnCode = part->confFn(interp, totargc-argc, argv+argc);" << endl;
   // confFn may set error code & error msg; just carry on
   tab--; of << tab << "}" << endl;
   
   of << tab << "return res;" << endl; 

   of << "}" << endl;

   of << endl;
}

   // classMethod

   of << "int " << name << OTCL_CLASS_SUFFIX << "::classMethod ("
      << "Tcl_Interp *i, int argc, char *argv[])" << endl;
   of << "{" << endl;
   // Make it like execute method
   of << tab << "int isfound = 0;" << endl;
   of << tab << "int *found = &isfound;" << endl;
   of << tab << "char *mName = argv[1];" << endl;
   of << endl;
   of << tab << "argc -= 2; argv += 2;" << endl;
       
   if(noOfClassSlots > 0)  {
       of << tab << "if(strcmp(mName, \"cget\")==0) { " << endl; tab++;
       of << tab << "return cgetFn(i, argc, argv);" << endl;
       tab--; of << tab << "}" << endl;
       of << tab << "if(strcmp(mName, \"config\")==0) { " << endl; tab++;
       of << tab << "return confFn(i, argc, argv);" << endl;
       tab--; of << tab << "}" << endl;
   }

   int m;
   for (m = 0; m < noOfClassMethods; m++)
   {
      classMethod[m]->executeBinding(name,cname,of,tab,this);
      of << endl;
   }
   of << tab << "Otcl::setTclResult(i,CLASS_METHOD_NOT_FOUND_ERR,"
      << "mName,\"" << name << "\", giveClassMethods());" << endl;
   of << tab << "return TCL_ERROR;" << endl;
   of << "}" << endl;
   of << endl;

   // giveClassMethods
   of << "char *" << name << OTCL_CLASS_SUFFIX << "::giveClassMethods (void)" << endl;
   of << "{" << endl;
   of << "   return \"" ;
{ 
    int i, first = 1;
    if (noOfClassSlots > 0)  {   // class only accepts these commands if it has slots
	of << "config/cget";
	first = 0;	// was 1, but now have "config/cget" prefix
    }
   for(i = 0; i < noOfClassMethods; ++i)  {
       if(!first)	of << "/";
       first = 0;
       of << classMethod[i]->giveName();
   }
}
   of << "\";" << endl;
   of << "}" << endl;

   if(noOfClassSlots > 0) {
       // cgetFn and confFn
       of << endl;
       of << "int " << name << OTCL_CLASS_SUFFIX << "::cgetFn (Tcl_Interp *i,"
	   << " int argc, char *argv[])" << endl;
       of << "{" << endl;
       of << tab << "int found = 0;" << endl;
       of << tab << "if(argc!=1)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"wrong # args: should be \\\"<class> cget -<slotname>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;
       of << tab << "if(argv[0][0]!='-')  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"bad slotname format: should be \\\"<class> cget -<slotname>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;			
       for (m = 0; m < noOfClassSlots; m++) {
	   if(m==0)	of << tab;
	   of << "if(strcmp(&argv[0][1], \""<< classSlot[m]->giveName() 
	       << "\")==0) {" << endl;	tab++;
	   of << tab << "found = 1;" << endl;
	   classSlot[m]->setReturn(name,of,tab);
	   tab--; of << tab << " } ";		
	   if(m<noOfClassSlots-1)
	       of << " else ";
	   else
	       of << endl;
       }
       of << tab << "if(found==0)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"no such classSlot \\\"\", &argv[0][1], \"\\\" for class \\\"" << name << "\\\" (not one of: ";
       int mm;
       for(mm=0; mm<noOfClassSlots; ++mm)  {
           of << classSlot[mm]->giveName();
           if(mm+1 < noOfClassSlots)	of << "/";
       }
       of << ")\", NULL);"<<endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;				
       
       of << tab << "return TCL_OK;" << endl;
       of << "}" << endl;

       of << endl;
       of << "int " << name << OTCL_CLASS_SUFFIX << "::confFn (Tcl_Interp *i,"
	   << " int argc, char *argv[])" << endl;
       of << "{" << endl;
       of << tab << "int found;" << endl;
       of << tab << "int ix = 0;" << endl;
       of << tab << "if(argc==0 || (argc&1)==1)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"wrong # args: should be \\\"<class> config -<slotname> <val>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;			
       of << tab << "while(ix < argc) {" << endl; tab++;
       of << tab << "found = 0;" << endl;
       of << tab << "if(argv[ix][0]!='-')  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"bad slotname format: should be \\\"<class> config -<slotname> <val>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;
       for (m = 0; m < noOfClassSlots; m++) {
	   if(m==0)	of << tab;
	   of << "if(strcmp(&argv[ix][1], \""<< classSlot[m]->giveName() 
	       << "\")==0) {" << endl;	tab++;
	   of << tab << "found = 1;" << endl;
	   classSlot[m]->setValFromStr(name,"argv[ix+1]",of,tab);

	   tab--; of << tab << " } ";		
	   if(m<noOfClassSlots-1)
	       of << " else ";
	   else
	       of << endl;
       }
       of << tab << "if(found==0)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"no such classSlot \\\"\", &argv[ix][1], \"\\\" for class \\\"" << name << "\\\" (not one of: ";
       for(mm=0; mm<noOfClassSlots; ++mm)  {
           of << classSlot[mm]->giveName();
           if(mm+1 < noOfClassSlots)	of << "/";
       }
       of << ")\", NULL);"<<endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;
       of << tab << "ix += 2;" << endl;
       tab--; of << "}" << endl;
       
       of << tab << "return TCL_OK;" << endl;
       of << "}" << endl;
   }
   of << endl;

}

void CdlClass::genOtclPartDec (ofstream &of)
{
    if(dummy)	return;

    if(castable)  {
	// does NOT inherit from C++ object
	of << "class " << name << OTCL_PART_SUFFIX << " : public OtclPartCpp" 
	    << endl;
    } else {
	// not copy constructed, allow it to be a true descendent of C++
	of << "class " << name << OTCL_PART_SUFFIX << " : public OtclPartCpp, "
	    << "public " << cname << endl;
    }
   of << "{" << endl;
   of << "private:" << endl;
   of << "   " << cname << " *objPtr;" << endl;	// companion C++ object
   of << "public:" << endl;
   // make fake slots public so I can use them to initialize constructor 
   // args from within the _class::instantiatePart (maybe just make 
   // the _class class a friend?)
   int i;
   for(i = 0; i < noOfFakeSlots; ++i) {
       fakeSlot[i]->genDec(of);
   }
   // ownObj used to be private, but now I want to be able to set it
   // after creating the object in the 'obj' return type (return 
   // obj by value)
   // I also want it for all objects, regardless of whether they are 
   // castable .. should propagate constraints through grammar, but too late..
//   if(castable)  {
       of << "   " << "int ownObj;" << endl;	// should we free it?
//   }

   if(constructor) {	// only define a constructor prototype if is 1
       constructor->genDec(of);
   }

   of << endl;

   // Copy constructor:
   of << "   " << name << OTCL_PART_SUFFIX << "(const " << cname << " &);"
      << endl;

   of << "   ~" << name << OTCL_PART_SUFFIX << " ();" << endl;

   int m;
   for (m = 0; m < noOfInstanceMethods; m++)
   {
      if (instanceMethod[m]->isDynamic())
      {
         instanceMethod[m]->genDec(of);
      }
   }

   of << "   char *giveClassName (void);" << endl;
   of << "   char *giveInstanceMethods (void);" << endl;	// dpwe
   of << "   char *giveSlots (void);" << endl;	// dpwe
   of << "   int *giveClassCountPtr (void);" << endl;	// dpwe
   of << "   int executeMethod (Tcl_Interp *, char *, int, char *[], int*);" 
      << endl;
   of << "   int discardPart (Tcl_Interp *, int);" << endl;
   of << "   void *toCpp (char *);" << endl;

   // Currently, always have cgetFn and confFn to allow ignorant passing 
   // upwards from subclasses
   if(1 || (noOfSlots+noOfFakeSlots) > 0)  {
    of << "   int cgetFn(Tcl_Interp *interp, int argc, char *argv[]);" << endl;
    of << "   int confFn(Tcl_Interp *interp, int argc, char *argv[]);" << endl;
   }
   of << "};" << endl;
}

void CdlClass::genOtclPartDef (ofstream &of)
{
   int i;

    if(dummy)	return;

if(constructor) {	// only define a constructor prototype if is 1

   // Constructor from OTCL instantiation
   of << name << OTCL_PART_SUFFIX << "::" << name << OTCL_PART_SUFFIX << " (";

   constructor->defineArgs(of);

   of << ") :" << endl;

   if(!castable)  {
       of << "   " << cname << "(";
       constructor->passArgsUpwards(of);
       of << ")," << endl;
   }

   of << "   OtclPartCpp(o)" << endl;
   of << "{" << endl;
   // either do a simple pointer assignment, or construct the C++ object
   if(castable)  {
//       of << "   objPtr = new " << cname << "("; 
//       constructor->passArgsUpwards(of);
//       of << ");" << endl;
//       of << "   ownObj = 1;" << endl;	// flag to destruct too 
//       of << "   EnsureOwner();" << endl;
//       of << "   " << name << OTCL_CLASS_SUFFIX 
//	  << "::pinstance->add_otcl(objPtr, this);" << endl;
       of << "   objPtr = NULL;" << endl;
   } else {
       // Hack - use the pointer, but point it to ourselves
       of << "   objPtr = (" << cname << " *)this;" << endl;
       // Do the deferred owner creation thing
       of << "   EnsureOwner();" << endl;
       of << "   " << name << OTCL_CLASS_SUFFIX 
	  << "::pinstance->add_otcl(objPtr, this);" << endl;
   }
   // initialize the fake slots
   for(i = 0; i < noOfFakeSlots; ++i) {
       fakeSlot[i]->genDefaultInit(of);
   }

   of << "}" << endl;

   of << endl;

   if(castable) {
       // Need a "newObject" method too to create object *after* shell
       of << "void " << name << OTCL_PART_SUFFIX << "::newObject(";
       constructor->defineArgs(of, 0);
       of << ")" << endl << "{" << endl;
       of << "   objPtr = new " << cname << "("; 
       constructor->passArgsUpwards(of);
       of << ");" << endl;
       of << "   ownObj = 1;" << endl;	// flag to destruct too 
       of << "   EnsureOwner();" << endl;
       of << "   " << name << OTCL_CLASS_SUFFIX 
	  << "::pinstance->add_otcl(objPtr, this);" << endl;
       of << "}" << endl;
       of << endl;
   }
}

   // Copy constructor
   of << name << OTCL_PART_SUFFIX << "::" << name << OTCL_PART_SUFFIX << " (";
   of << "const " << cname << " &obj) :" << endl;
   if(!castable)  {
       of << "   " << cname << "(obj)," << endl;
   }
   of << "    OtclPartCpp(NULL)" << endl;
   of << "{" << endl;
   if(castable)  {
       // have to cast to get around const-ness?
       of << "   objPtr = (" << cname << " *)&obj;" << endl;
       of << "   ownObj = 0;" << endl;	// did NOT construct it here
   } else {
       of << "   objPtr = (" << cname << " *)this;" << endl;
   }
   // Do the deferred owner creation thing
   of << "   EnsureOwner();" << endl;
   of << "   " << name << OTCL_CLASS_SUFFIX 
       << "::pinstance->add_otcl(objPtr, this);" << endl;
   // Make sure the dummy slots are at least init'ed, even though 
   // values probably won't reflect contents of acquired object
   for(i = 0; i < noOfFakeSlots; ++i) {
       fakeSlot[i]->genDefaultInit(of);
   }
   of << "}" << endl;

   of << endl;

   // Destructor
   of << name << OTCL_PART_SUFFIX << "::~" << name << OTCL_PART_SUFFIX 
      << "()" << endl;
   of << "{" << endl;
   of << "   " << name << OTCL_CLASS_SUFFIX 
       << "::pinstance->remove_otcl(objPtr);" << endl;
   of << "   if (!owner->discardingParts())" << endl;
   of << "   {" << endl;
   of << "      Otcl::otclPtr->discard(Otcl::tclInterp,owner->getSelf(),"
      << "OTCL_TRUE);" << endl;
   of << "   }" << endl;
   if(castable)  {
       of << "   if(ownObj)  {" << endl << "      delete objPtr;" << endl 
	   << "   }" << endl;
   }
   for(i = 0; i < noOfFakeSlots; ++i) {
       fakeSlot[i]->genDestruct(of);
   }
   of << "}" << endl;

   of << endl;

   // giveClassName
   of << "char *" << name << OTCL_PART_SUFFIX << "::giveClassName (void)" << endl;
   of << "{" << endl;
   of << "   return \"" << name << "\";" << endl;
   of << "}" << endl;

   of << endl;

   // giveInstanceMethods
   of << "char *" << name << OTCL_PART_SUFFIX << "::giveInstanceMethods (void)" << endl;
   of << "{" << endl;
   of << "   return \"" ;
{ 
    int i, first = 1;
    if ((noOfSlots+noOfFakeSlots) > 0)  {   // obj only accepts these commands if it has slots
	of << "config/cget";
	first = 0;	// was 1, but now have "config/cget" prefix
    }
   for(i = 0; i < noOfInstanceMethods; ++i)  {
       if(!first)	of << "/";
       first = 0;
       of << instanceMethod[i]->giveName();
   }
}
   of << "\";" << endl;
   of << "}" << endl;

   of << endl;

   // giveSlots
   of << "char *" << name << OTCL_PART_SUFFIX << "::giveSlots (void)" << endl;
   of << "{" << endl;
   of << "   return \"" ;
{ 
    int i, first = 1;
   for(i = 0; i < noOfSlots; ++i)  {
       if(!first)	of << "/";
       first = 0;
       of << slot[i]->giveName();
   }
   for(i = 0; i < noOfFakeSlots; ++i)  {
       if(!first)	of << "/";
       first = 0;
       of << fakeSlot[i]->giveName();
   }
}
   of << "\";" << endl;
   of << "}" << endl;

   of << endl;

   // giveClassCountPtr
   of << "int *" << name << OTCL_PART_SUFFIX << "::giveClassCountPtr (void)" << endl;
   of << "{" << endl;
   of << "   return " << name << OTCL_CLASS_SUFFIX << "::pinstance->giveCountPtr();" << endl;
   of << "}" << endl;

   of << endl;

   // discardPart
   of << "int " << name << OTCL_PART_SUFFIX
      << "::discardPart (Tcl_Interp *, int fromCpp)" << endl;
   of << "{" << endl;
   of << "   if (fromCpp == OTCL_TRUE) return TCL_OK;" << endl;
   of << "   delete this;" << endl;
   of << "   return TCL_OK;" << endl;
   of << "}" << endl;

   of << endl;

   // toCpp
   of << "void *" << name << OTCL_PART_SUFFIX
      << "::toCpp (char *cppClassName)" << endl;
   of << "{" << endl;

   of << "   if (strcmp(cppClassName,\"" << name << "\") == 0)" << endl;
   of << "   {" << endl;
//   of << "      return (void*)((" << cname << "*)this);" << endl;
   of << "      return (void*)objPtr;" << endl;
   of << "   }" << endl;

   int s;
   for (s = 0; s < noOfSuperclasses; s++)
   {
      of << "   if (strcmp(cppClassName,\"" << superclasses[s] << "\") == 0)"
         << endl;
      of << "   {" << endl;
//     of << "      return (void*)((" << superclasses[s] << "*)this);" << endl;
      of << "      return (void*)((" << superclasses[s] << "*)objPtr);" <<endl;
      of << "   }" << endl;
   } 

   of << "   return NULL;" << endl;
   of << "}" << endl;

   of << endl;

   // executeMethod
   of << "int " << name << OTCL_PART_SUFFIX << "::executeMethod (Tcl_Interp *i,"
      << " char *mName, int argc," << endl;
   of << "         char *argv[], int *found)" << endl;
   of << "{" << endl;

   CdlIndent tab(1);

   if((noOfSlots+noOfFakeSlots) > 0)  {
       of << tab << "if(strcmp(mName, \"cget\")==0) { " << endl; tab++;
       of << tab << "*found=OTCL_TRUE; return cgetFn(i, argc, argv);" << endl;
       tab--; of << tab << "}" << endl;
       of << tab << "if(strcmp(mName, \"config\")==0) { " << endl; tab++;
       of << tab << "*found=OTCL_TRUE; return confFn(i, argc, argv);" << endl;
       tab--; of << tab << "}" << endl;
   }

   int m;
   for (m = 0; m < noOfInstanceMethods; m++)
   {
      instanceMethod[m]->executeBinding(name, cname, of, tab, this);
      of << endl;
   }

   // May withdraw these 'give_<slotname>' methods now we have 'cget -<slot>'
//   for (m = 0; m < noOfSlots; m++)
//   {
//      slot[m]->executeBinding(name,of);
//      of << endl;
//   }

   // Sort-of hack.  Call the execute methods of statically-defined 
   // superclasses to see if any methods inherit.  Experimental 1995nov29.
   // You can only do this if the class is castable, since we're going 
   // to have to generate a new OTCL object on the fly for the parent 
   // class - can't do that if we're inheriting, only if we're adjunct
   if(castable)  {
       for(m=0; m<noOfSuperclasses; ++m)  {
	   // Construct an object which is the _otcl for this object as
	   // its superclass
	   // "  SCLASS_otcl* this_SCLASS = (SCLASS_otcl*)
	   //        SCLASS_otclc::instance.find_otcl((SCLASS*)objPtr); "
	   char *sclass = superclasses[m];
           of << "   " << sclass << "_otcl* this_" << sclass << " = ";
	   of << "(" << sclass << "_otcl*)";
	   of << sclass << "_otclc::pinstance->find_otcl(";
	   of << "(" << sclass << "*)objPtr);" << endl;
	   of << "   this_" << superclasses[m] << "->executeMethod(i, mName, argc, argv, found);" << endl;
	   of << "   if(*found == OTCL_TRUE)  return TCL_OK;" << endl;
       }
   }

   of << "   *found = OTCL_FALSE;" << endl;
   of << "   return TCL_OK;" << endl;
   of << "}" << endl;

   // Dynamic method definitions
   for (m = 0; m < noOfInstanceMethods; m++)
   {
      if (instanceMethod[m]->isDynamic())
      {
         of << endl;
         instanceMethod[m]->genDef(name,cname,of,castable);
      }
   }

   // Have to *always* generate cgetFn and confFn, even if there are 
   // no slots, so that upwards passing in subclasses will have something 
   // to call, even if it's a waste of time
   if(1 || (noOfSlots+noOfFakeSlots) > 0) {
       // cgetFn and confFn
       of << endl;
       of << "int " << name << OTCL_PART_SUFFIX << "::cgetFn (Tcl_Interp *i,"
	   << " int argc, char *argv[])" << endl;
       of << "{" << endl;
       // CdlIndent tab(1);
       of << tab << "int found = 0;" << endl;
       of << tab << "if(argc!=1)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"wrong # args: should be \\\"<object> cget -<slotname>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;				
       of << tab << "if(argv[0][0]!='-')  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"bad slotname format: should be \\\"<object> cget -<slotname>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;			
       for (m = 0; m < noOfSlots; m++) {
	   if(m==0)	of << tab;
	   of << "if(strcmp(&argv[0][1], \""<< slot[m]->giveName() 
	       << "\")==0) {" << endl;	tab++;
	   of << tab << "found = 1;" << endl;
	   slot[m]->setReturn(of,tab);
	   tab--; of << tab << " } ";		
	   if(m<noOfSlots-1)
	       of << " else ";
	   else
	       of << endl;
       }
       for (m = 0; m < noOfFakeSlots; m++) {
	   if(m==0)	of << tab;
	   of << "if(strcmp(&argv[0][1], \""<< fakeSlot[m]->giveName() 
	       << "\")==0) {" << endl;	tab++;
	   of << tab << "found = 1;" << endl;
	   if(fakeSlot[m]->getScript != NULL && strlen(fakeSlot[m]->getScript)) {
	       // Execute getScript before returning
	       // setup local var frame
	       of << tab << "Otcl::createProcScope(i);" << endl;
	       // initialize local variables: 'this'
	       of << tab << "Tcl_SetVar(i,\"" << TCL_THIS << "\",owner->getSelf(),0);" << endl;
	       // .. and other dummy vars
	       of << tab << "char *str;" << endl;
	       of << tab << "int rc;" << endl;
	       int i;
	       for(i = 0; i<noOfFakeSlots; ++i) {
		   fakeSlot[i]->initString("str", of, tab);
		   of << tab << "Tcl_SetVar(i, \"" << fakeSlot[i]->giveName() << "\", str, 0);" << endl; 
		   fakeSlot[i]->cleanString("str", of, tab);
	       }
	       // run the script
       // Under DEC-OSF1, the temporary in-place modifications to the 
       // string being eval'd (in Tcl_ParseVar) cause a seg fault; 
       // therefore, we copy the static command string into a dynamic
       // buffer before eval'ing (1997aug10)
       of << tab << "str = strdup(\"";
	       putScript(of, fakeSlot[m]->getScript);
       of << "\");" << endl;
       of << tab << "rc = Tcl_Eval(i, str);" << endl;
       of << tab << "free(str);" << endl;
	       of << tab << "if( rc == TCL_ERROR)  return TCL_ERROR;" << endl;
	       // set the dummy variable from the Tcl scope
	       char expr[64];
	       sprintf(expr, "Tcl_GetVar(i, \"%s\", 0)", fakeSlot[m]->giveName());
	       fakeSlot[m]->setValFromStr(expr,of,tab);
	       // detach the stack frame
	       of << tab << "Otcl::destroyProcScope(i);" << endl;
	   }
	   fakeSlot[m]->setReturn(of,tab);
	   tab--; of << tab << " } ";		
	   if(m<noOfFakeSlots-1)
	       of << " else ";
	   else
	       of << endl;
       }
       // Optionally pass up to cgetFns of parent classes - if a castable type
       if(castable)  {
	   for(m=0; m<noOfSuperclasses; ++m)  {
	       of << tab << "if(found==0)  {" << endl;	tab++;
	       // Construct an object which is the _otcl for this object as
	       // its superclass
	       // "  SCLASS_otcl* this_SCLASS = (SCLASS_otcl*)
	       //        SCLASS_otclc::instance.find_otcl((SCLASS*)objPtr); "
	       char *sclass = superclasses[m];
	       of << tab << sclass << "_otcl* this_" << sclass << " = ";
	       of << "(" << sclass << "_otcl*)";
	       of << sclass << "_otclc::pinstance->find_otcl(";
	       of << "(" << sclass << "*)objPtr);" << endl;
	       of << tab << "if(this_" << superclasses[m] << "->cgetFn(i, argc, argv)==TCL_OK) found = 1; else Tcl_AppendResult(i, \"\\n\", NULL);" << endl;
	       tab--; of << tab << "}" << endl;
	   }
       }
       of << tab << "if(found==0)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"no such slot \\\"\", &argv[0][1], \"\\\" for class \\\"" << name << "\\\" (not one of: ";
       int mm;
       int first = 1;
       for(mm=0; mm<noOfSlots; ++mm)  {
	   if (!first) {of << "/";} else {first = 0;}
           of << slot[mm]->giveName();
       }
       for(mm=0; mm<noOfFakeSlots; ++mm)  {
	   if (!first) {of << "/";} else {first = 0;}
           of << fakeSlot[mm]->giveName();
       }
       of << ")\", NULL);"<<endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;				
       
       of << tab << "return TCL_OK;" << endl;
       of << "}" << endl;

       of << endl;
       of << "int " << name << OTCL_PART_SUFFIX << "::confFn (Tcl_Interp *i,"
	   << " int argc, char *argv[])" << endl;
       of << "{" << endl;
       of << tab << "int found;" << endl;
       of << tab << "int ix = 0;" << endl;
       of << tab << "if(argc==0 || (argc&1)==1)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"wrong # args: should be \\\"<object> config -<slotname> <val>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;
       of << tab << "while (ix < argc) {" << endl; tab++;
       of << tab << "found = 0;" << endl;
       of << tab << "if(argv[ix][0]!='-')  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"bad slotname format: should be \\\"<object> config -<slotname> <val>\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;					
       for (m = 0; m < noOfSlots; m++) {
	   if(m==0)	of << tab;
	   of << "if(strcmp(&argv[ix][1], \""<< slot[m]->giveName() 
	       << "\")==0) {" << endl;	tab++;
	   of << tab << "found = 1;" << endl;
	   slot[m]->setValFromStr("argv[ix+1]",of,tab);

	   tab--; of << tab << " } ";
	   if(m<noOfSlots-1)
	       of << " else ";
	   else
	       of << endl;
       }
       for (m = 0; m < noOfFakeSlots; m++) {
	   if(m==0)	of << tab;
	   of << "if(strcmp(&argv[ix][1], \""<< fakeSlot[m]->giveName() 
	       << "\")==0) {" << endl;	tab++;
	   of << tab << "found = 1;" << endl;
	   fakeSlot[m]->setValFromStr("argv[ix+1]",of,tab);

	   if(fakeSlot[m]->setScript != NULL && strlen(fakeSlot[m]->setScript)) {
	       // Execute setScript before returning - but only if object 
	       // exists i.e. this isn't the pre-constructor call to confFn
	       if(castable) {
		   of << tab << "if (objPtr) {" << endl; tab++;
	       }
	       // setup local var frame
	       of << tab << "Otcl::createProcScope(i);" << endl;
	       // initialize local variables: 'this'
	       of << tab << "Tcl_SetVar(i,\"" << TCL_THIS << "\",owner->getSelf(),0);" << endl;
	       // .. and other dummy vars
	       of << tab << "char *str;" << endl;
	       of << tab << "int rc;" << endl;
	       int i;
	       for(i = 0; i<noOfFakeSlots; ++i) {
		   fakeSlot[i]->initString("str", of, tab);
		   of << tab << "Tcl_SetVar(i, \"" << fakeSlot[i]->giveName() << "\", str, 0);" << endl; 
		   fakeSlot[i]->cleanString("str", of, tab);
	       }
	       // run the script
       // Under DEC-OSF1, the temporary in-place modifications to the 
       // string being eval'd (in Tcl_ParseVar) cause a seg fault; 
       // therefore, we copy the static command string into a dynamic
       // buffer before eval'ing (1997aug10)
       of << tab << "str = strdup(\"";
	       putScript(of, fakeSlot[m]->setScript);
       of << "\");" << endl;
       of << tab << "rc = Tcl_Eval(i, str);" << endl;
       of << tab << "free(str);" << endl;
	       of << tab << "if( rc == TCL_ERROR)  return TCL_ERROR;" << endl;
	       // detach the stack frame
	       of << tab << "Otcl::destroyProcScope(i);" << endl;
	       if(castable) {
		   // close the "if (objPtr)" clause
		   tab--; of << tab << "}" << endl;
	       }
	   }
	   tab--; of << tab << " } ";
	   if(m<noOfFakeSlots-1)
	       of << " else ";
	   else
	       of << endl;
       }
       // Optionally pass up to confFns of parent classes - if a castable type
       if(castable)  {
	   for(m=0; m<noOfSuperclasses; ++m)  {
	       of << tab << "if(found==0)  {" << endl;	tab++;
	       // Construct an object which is the _otcl for this object as
	       // its superclass
	       // "  SCLASS_otcl* this_SCLASS = (SCLASS_otcl*)
	       //        SCLASS_otclc::pinstance->find_otcl((SCLASS*)objPtr); "
	       char *sclass = superclasses[m];
	       of << tab << sclass << "_otcl* this_" << sclass << " = ";
	       of << "(" << sclass << "_otcl*)";
	       of << sclass << "_otclc::pinstance->find_otcl(";
	       of << "(" << sclass << "*)objPtr);" << endl;
	       of << tab << "if(this_" << superclasses[m] << "->confFn(i, 2, argv+ix)==TCL_OK) found = 1; else Tcl_AppendResult(i, \"\\n\", NULL);" << endl;
	       tab--; of << tab << "}" << endl;
	   }
       }
       of << tab << "if(found==0)  {" << endl;	tab++;
       of << tab << "Tcl_AppendResult(i, \"no such slot \\\"\", &argv[ix][1], \"\\\" for class \\\"" << name << "\\\" (not one of: ";
       first = 1;
       for(mm=0; mm<noOfSlots; ++mm)  {
	   if (first) {first = 0;} else {of << "/";}
           of << slot[mm]->giveName();
       }
       for(mm=0; mm<noOfFakeSlots; ++mm)  {
	   if (first) {first = 0;} else {of << "/";}
           of << fakeSlot[mm]->giveName();
       }
       of << ")\", NULL);"<<endl;
       of << tab << "return TCL_ERROR;" << endl;
       tab--; of << tab << "}" << endl;				
       
       of << tab << "ix += 2;" << endl;
       tab--; of << tab << "}" << endl;

       of << tab << "return TCL_OK;" << endl;
       of << "}" << endl;
   }
}

void CdlClass::addClass(char *name, CdlClass *c) {
    classlist = new classRecdNode(classlist, name, c);
//    DBGFPRINTF((stderr, "**addClass: added %s\n", name));
}
classRecdNode **CdlClass::findClassNodeAdr(char *name) {
    classRecdNode **psrch = &classlist;
//    DBGFPRINTF((stderr, "**findClass: for %s .. \n", name));
    while(*psrch != NULL) {
//	DBGFPRINTF((stderr, "**findClass: is it %s? ", (*psrch)->name));
	if(strcmp(name, (*psrch)->name)==0) {
//	    DBGFPRINTF((stderr, "yes\n"));
	    return psrch;
	} else {
	    psrch = &((*psrch)->next);
//	    DBGFPRINTF((stderr, "no\n"));
	}
    }
    return psrch;
}
void CdlClass::rmvClass(char *name) {
    classRecdNode **precd = findClassNodeAdr(name);
    if (precd != NULL && *precd != NULL) {
	*precd = (*precd)->next;
	delete *precd;
    } else {
	fprintf(stderr, "CdlClass::rmvClass: can't find %s\n", name);
    }
}
CdlClass *CdlClass::findClass(char *name) {
    classRecdNode **precd = findClassNodeAdr(name);
    if (precd && *precd) {
	return (*precd)->classPtr;
    } else {
	return NULL;
    }
}
