/*  _ __ ___ _
 * | |\ /  /| |  $Id: CdlMethod.C,v 1.8 1997/10/10 23:10:44 dpwe Exp $
 * | | /  / | |  Copyright (C) 1995 IXI Limited.
 * |_|/__/_\|_|  IXI Limited, Cambridge, England.
 *
 * Component   : CdlMethod.C
 *
 * Author      : Dean Sheehan (deans@x.co.uk)
 *
 * Description : Contains the implementation for CdlMethod class which modesl
 *               methods on CDL'ed classes. Also include implementation for
 *               various specialised method subclasses like constructors.
 *
 * License     :
			Object Tcl License & Copyright
			------------------------------

IXI Object Tcl software, both binary and source (hereafter, Software) is copyrighted by IXI Limited (IXI), and ownership remains with IXI. 

IXI grants you (herafter, Licensee) a license to use the Software for academic, research and internal business purposes only, without a fee. Licensee may distribute the binary and source code (if required) to third parties provided that the copyright notice and this statement appears on all copies and that no charge is associated with such copies. 

Licensee may make derivative works. However, if Licensee distributes any derivative work based on or derived from the Software, then Licensee will (1) notify IXI regarding its distribution of the derivative work, and (2) clearly notify users that such derivative work is a modified version and not the original IXI Object Tcl distributed by IXI. IXI strongly recommends that Licensee provide IXI the right to incorporate such modifications into future releases of the Software under these license terms. 

Any Licensee wishing to make commercial use of the Software should contact IXI, to negotiate an appropriate license for such commercial use. Commercial use includes (1) integration of all or part of the source code into a product for sale or license by or on behalf of Licensee to third parties, or (2) distribution of the binary code or source code to third parties that need it to utilize a commercial product sold or licensed by or on behalf of Licensee. 

IXI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY. IXI SHALL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER SUFFERED BY THE USERS OF THIS SOFTWARE. 

Copyright (C) 1995, IXI Limited 

By using or copying this Software, Licensee agrees to abide by the copyright law and all other applicable laws of England and the U.S., including, but not limited to, export control laws, and the terms of this license. IXI shall have the right to terminate this license immediately by written notice upon Licensee's breach of, or non-compliance with, any of its terms. Licensee may be held legally responsible for any copyright infringement that is caused or encouraged by Licensee's failure to abide by the terms of this license. 

Comments and questions are welcome and can be sent to
otcl@x.co.uk 

For more information on copyright and licensing issues, contact: 
Legal Department, IXI Limited, Vision Park, Cambridge CB4 4ZR,
ENGLAND. 

 *
 */

// System Includes
#include <stdlib.h>
#include <string.h>
#include <assert.h>

// Local Includes
#include "CdlFile.H"
#include "CdlMethod.H"
#include "CdlArg.H"
#include "CdlRtn.H"

#include "CdlSlot.H"

void putScript(ostream& of, char *script)
{    // Global fn to quote script strings - by escaping double-quotes!
    char c;
    while( (c = *script++) != '\0') {
	if ( c == '"') {
	    of << '\\';
	}
	of << c;
    }
}

CdlMethod::CdlMethod (char *n)
{
   name = strdup(n);
   noOfArgs = 0;
   minNoOfArgs = 0;
   voidReturn = 0;
   voidArgs = 0;
   rtn = NULL;
   script = NULL;
   binding = STATIC;
}

CdlMethod::~CdlMethod ()
{
   if (rtn != NULL)
   {
      delete rtn;
   }
   int a;
   for (a = 0; a < noOfArgs; a++)
   {
      delete arg[a];
   }
   if(script != NULL)	free(script);
}

void CdlMethod::setScript(const char *str)
{
    script = strdup(str);
}

void CdlMethod::setVoidArgs (void)
{
   voidArgs = 1;
}

int CdlMethod::argsAreVoid (void)
{
   return voidArgs;
}

void CdlMethod::setVoidReturn (void)
{
   voidReturn = 1;
}

int CdlMethod::returnIsVoid (void)
{
   return voidReturn;
}

int CdlMethod::addArg (CdlArg *a)
{
    if(noOfArgs > minNoOfArgs) {
	// Preceding arg was optional, so this one had better be too
	if(a->defaultFlag == ARG_DEFAULT_NONE) {
	    return 0;
	}
    }
    arg[noOfArgs] = a;
    noOfArgs++;
    if(a->defaultFlag == ARG_DEFAULT_NONE) {
	minNoOfArgs++;
    }
    return 1;
}

void CdlMethod::setRtn (CdlRtn *r)
{
   rtn = r;
}

CdlRtn *CdlMethod::getRtn (void)
{
   return rtn;
}

int CdlMethod::getNoOfArgs (void)
{
   return noOfArgs;
}

int CdlMethod::getMinNoOfArgs (void)
{
   return minNoOfArgs;
}

CdlInstanceMethod::CdlInstanceMethod (char *n) :
   CdlMethod(n)
{
}

CdlInstanceMethod::~CdlInstanceMethod ()
{
}

void CdlMethod::setBinding (Binding b)
{
   binding = b;
}

void CdlInstanceMethod::genDec (ofstream &of)
{
   of << "   ";

   if (voidReturn)
   {
      of << "void ";
   }
   else
   {
      rtn->genPrototype(of);
   }

   of << name << " (";

   int a;
   for (a = 0; a < noOfArgs; a++)
   {
      arg[a]->type(of);
      if (a < (noOfArgs-1))
      {
         of << ", ";
      }
   }
   of << ");" << endl;
}

int CdlMethod::isDynamic (void)
{
   return (binding == DYNAMIC ? 1 : 0);
}

void CdlInstanceMethod::executeBinding (char *className, char *cname, 
					ofstream &of, CdlIndent tab, 
					CdlClass* pclass)
{
    executeBinding_sub(className, cname, of, tab, pclass, "objPtr->");
}

void CdlMethod::executeBinding_sub (char *className, char *cname, 
					ofstream &of, CdlIndent tab, 
					CdlClass* pclass, 
					    char *cmdPrefix)
{   // part common to method and classMethod
    // for method, cmdPrefix = "objPtr->" ;
    // for classMethod, it's "ClassName::"
   char varName[32];
   char strName[32];

   of << tab << "if (strcmp(mName,\"" << name << "\") == 0)" << endl;
   of << tab << "{" << endl;    tab++;
   of << tab << "int rc = TCL_OK;" << endl;

   of << tab << "*found = OTCL_TRUE;" << endl;

   if(binding == DUMMYCPP) {
       // special flag that the script is C code, not Tcl
       int a;
       // Check number of args
       of << tab << "if(argc<" << minNoOfArgs << " || argc>" << noOfArgs << ") {" << endl; tab++;
       of << tab << "Tcl_AppendResult(i, \"wrong # args: should be: \\\" className :: \", mName";
       for (a = 0; a < minNoOfArgs; a++) {
	   of << ", \" " << arg[a]->giveCommand() << "\"";
       }
       for (a = minNoOfArgs; a < noOfArgs; a++) {
	   of << ", \" ?" << arg[a]->giveCommand() << "?\"";
       }
       of << ", \"\\\"\", NULL);" << endl;
       of << tab << "return TCL_ERROR;" << endl;
       of << tab << "}" << endl; tab--;
       // Set up the args
       for (a = 0; a < minNoOfArgs; a++)   {
	   sprintf(varName,"arg%d",a);
	   sprintf(strName,"argv[%d]",a);
	   arg[a]->initVarFromStr(varName,strName,of,tab,1);
       }
       for (a = minNoOfArgs; a < noOfArgs; a++) {
	   sprintf(varName,"arg%d",a);
	   if (arg[a]->defaultFlag == ARG_DEFAULT_VAL) {
	       sprintf(strName,"\"%s\"", arg[a]->defaultVal);
	       arg[a]->initVarFromStr(varName,strName,of,tab,1);
	   } else {
	       assert(arg[a]->defaultFlag == ARG_DEFAULT_SLOT);
	       arg[a]->initFromVar(varName,arg[a]->defaultVal,of,tab,1);
	   }
	   of << tab << "if(argc > " << a << ") {" << endl; tab++;
	   sprintf(strName,"argv[%d]",a);
	   arg[a]->initVarFromStr(varName,strName,of,tab,1,1);
	   tab--; of << tab << "}" << endl;
       }
       if (!voidReturn) {
	   rtn->declareLocal(of,tab);
	   /* hack - don't set rtn yet, let the code do it */
	   of << "0);" << endl;
       }
   
       // write out the script
       of << "{" << script;
       if (!voidReturn) {
	   rtn->postProcess(of,tab);
       }
       of << "}" << endl;

       for (a = 0; a < noOfArgs; a++) {
	   sprintf(varName,"arg%d",a);
	   sprintf(strName,"argv[%d]",a);
	   arg[a]->cleanVarFromStr(varName,strName,of,tab);
       }
   } else if(binding == DUMMYTCL) {
       // setup local var frame
       of << tab << "Otcl::createProcScope(i);" << endl;
       // initialize 'args'
       of << tab << "int ii; Tcl_DString ds; Tcl_DStringInit(&ds);" << endl;
       of << tab << "for (ii = 0; ii < argc; ++ii) {Tcl_DStringAppendElement(&ds, argv[ii]);}" << endl;
       of << tab << "Tcl_SetVar(i,\"args\",Tcl_DStringValue(&ds),0);" << endl;
       of << tab << "Tcl_DStringFree(&ds);" << endl;
       // initialize local variables: 'this'
       of << tab << "Tcl_SetVar(i,\"" << TCL_THIS << "\",owner->getSelf(),0);" << endl;
       // .. and other dummy vars
       of << tab << "char *str;" << endl;
       int i;
       for(i = 0; i<pclass->noOfFakeSlots; ++i) {
	   pclass->fakeSlot[i]->initString("str", of, tab);
	   of << tab << "Tcl_SetVar(i, \"" << pclass->fakeSlot[i]->giveName() << "\", str, 0);" << endl; 
	   pclass->fakeSlot[i]->cleanString("str", of, tab);
       }
       // run the script
       // Under DEC-OSF1, the temporary in-place modifications to the 
       // string being eval'd (in Tcl_ParseVar) cause a seg fault; 
       // therefore, we copy the static command string into a dynamic
       // buffer before eval'ing (1997aug10)
       of << tab << "str = strdup(\"";
       putScript(of, script);
       of << "\");" << endl;
       of << tab << "rc = Tcl_Eval(i, str);" << endl;
       of << tab << "free(str);" << endl;
       of << tab << "if (rc != TCL_ERROR)	rc = TCL_OK;" << endl;
       // detach the stack frame
       of << tab << "Otcl::destroyProcScope(i);" << endl;
   } else {
       // conventional (static or dynamic) binding
//   of << tab << "if (argc != " << noOfArgs << ")" << endl;
       of << tab << "if (argc < " << minNoOfArgs << " || argc > " << noOfArgs << ")" << endl;
       of << tab << "{" << endl;
       tab++;
       of << tab << "return Otcl::setTclError(i,ARGS_METHOD_ERR,"
	  << "\"" << name << "\",\"" << className << "\");" << endl;
       tab--;
       of << tab << "}" << endl;

       int a;
       for (a = 0; a < minNoOfArgs; a++)   {
	   sprintf(varName,"arg%d",a);
	   sprintf(strName,"argv[%d]",a);
	   arg[a]->initVarFromStr(varName,strName,of,tab,1);
       }
       for (a = minNoOfArgs; a < noOfArgs; a++) {
	   sprintf(varName,"arg%d",a);
	   if (arg[a]->defaultFlag == ARG_DEFAULT_VAL) {
	       sprintf(strName,"\"%s\"", arg[a]->defaultVal);
	       arg[a]->initVarFromStr(varName,strName,of,tab,1);
	   } else {
	       assert(arg[a]->defaultFlag == ARG_DEFAULT_SLOT);
	       arg[a]->initFromVar(varName,arg[a]->defaultVal,of,tab,1);
	   }
	   of << tab << "if(argc > " << a << ") {" << endl; tab++;
	   sprintf(strName,"argv[%d]",a);
	   arg[a]->initVarFromStr(varName,strName,of,tab,1,1);
	   tab--; of << tab << "}" << endl;
       }
       if (!voidReturn) {
	   rtn->declareLocal(of,tab);
       } else {
	   of << tab << "(void)(";
       }
   
//   of << className << "::" << name << "(";
       if(isDynamic())  {
	   // for 'dynamic' (Otcl-subclassable) objects, have to 
	   // execute our class method to allow preemption
	   of << name << "(";
       } else {
	   // tcl pre-emption not allowed, just execute directly on object
	   of << cmdPrefix << name << "(";
       }
       for (a = 0; a < noOfArgs; a++) {
	   sprintf(varName,"arg%d",a);
	   of << varName;
	   if (a < (noOfArgs - 1)) {
	       of << ", ";
	   }
       }
       of << "));" << endl;	// second close paren new 1996jan25 - decLocal opens
       for (a = 0; a < noOfArgs; a++) {
	   sprintf(varName,"arg%d",a);
	   sprintf(strName,"argv[%d]",a);
	   arg[a]->cleanVarFromStr(varName,strName,of,tab);
       }

       if (!voidReturn) {
	   rtn->postProcess(of,tab);
       }
   }
   of << tab << "return rc;" << endl;
   tab--;
   of << tab << "}" << endl;
}

void CdlInstanceMethod::genDef (char *className, char *cName, ofstream &of, int castable)
{
   int a;
   char varName[32];
   char strName[32];

   CdlIndent tab(1);

   if (voidReturn)
   {  
      of << "void";
   }
   else
   {
      rtn->genPrototype(of);
   }

   of << " " << className << OTCL_PART_SUFFIX << "::" << name << "(";
   if (voidArgs)
   {
      of << "void";
   }
   else
   {
      for (a = 0; a < noOfArgs; a++) 
      {
         sprintf(varName,"arg%d",a);
         arg[a]->type(of);
         of << " " << varName;
         if (a < (noOfArgs-1))
         {
             of << ", ";
         }
      }
   }
   of << ")" << endl;
   of << "{" << endl;
   of << tab << "if (owner->giveOtclPart() == this)" << endl;
   of << tab << "{" << endl;

   tab++;

   if (voidReturn)
   {
      of << tab;
   }
   else
   {
      of << tab << "return ";
   }

//   of << cName << "::" << name << "(";
// If this object is derived from the C++ prototype, (i.e. !castable) 
// we *have* to specify the scope of the method to run, since if it is 
// virtual, this function will be the virtual method and we would just loop.
// Thus, we specify the C++ object's one, but we lose on any vtable definition
// on the object we are based on.
   if(!castable)  {
       of << "objPtr->" << cName << "::" << name << "(";
   } else {
       of << "objPtr->" << name << "(";
   }
   for (a = 0; a < noOfArgs; a++)
   {
      sprintf(varName,"arg%d",a);
      of << varName;
      if (a < (noOfArgs-1))
      {
         of << ", ";
      }
   }
   of << ");" << endl;
   if (voidReturn)
   {
      of << tab << "return;" << endl;
   }
   tab--;
   of << tab << "}" << endl;

   if (noOfArgs != 0)
   {
      of << tab << "char *argv[" << noOfArgs << "];" << endl;
      for (a = 0; a < noOfArgs; a++)
      {
         sprintf(varName,"arg%d",a);
         sprintf(strName,"argv[%d]",a);
         arg[a]->initStrFromVar(strName,varName,of,tab);
      }
   }

   of << endl;

   of << tab << "int result = owner->executeMethod(Otcl::tclInterp,\"" << name
      << "\"," << noOfArgs << ",";
   if (noOfArgs == 0)
   {
      of << "NULL";
   }
   else
   {
      of << "argv";
   }
   of << ");" << endl;

   of << endl;

   of << tab << "if (result != TCL_OK)" << endl;
   of << tab << "{" << endl;
   tab++;
   of << tab << "otclErrorMethod(\"" << className << "::" << name << "\",Otcl::tclInterp->result);" << endl;
   tab--;
   of << tab << "}" << endl;

   for (a = 0; a < noOfArgs; a++)
   {
      sprintf(varName,"arg%d",a);
      sprintf(strName,"argv[%d]",a);
      arg[a]->cleanStrFromVar(strName,varName,of,tab);
   }

   if (!voidReturn)
   {
     rtn->returnFromTclInterp("Otcl::tclInterp",of,tab);
   }

   of << "}" << endl;
}

CdlConstructor::CdlConstructor (char *cname, int castb /* = 0 */) :
   CdlMethod(cname)
{
    castable = castb;
}

CdlConstructor::~CdlConstructor ()
{
}

void CdlConstructor::genDec (ofstream &of)
{
   of << "   " << name << "(";

   if(!castable) {
       int a;
       for (a = 0; a < noOfArgs; a++) {
	   arg[a]->type(of);
	   of << ", ";
       }
   }
   of << "OtclObject *o = NULL);" << endl;
   if(castable) {
       int a;
       of << "    void newObject("; 
       for (a = 0; a < noOfArgs; a++) {
	   if(a)	of << ", ";
	   arg[a]->type(of);
       }
       of << ");" << endl;
   }
}

void CdlConstructor::defineArgs (ofstream &of, int includeOtcl /* = 1 */)
{   // Hack.  Use this to generate the arg lists both for the derived
    // _otcl part constructor (which takes the object constructor args
    // followed by an OtclObject) AND the *castable* (non-derived) 
    // constructor (which *doesn't* create the object, so doesn't take
    // any args) AND the castable's newObject method (which creates the 
    // object, but doesn't deal with any OtclObject).
    // this->castable = 0, includeOtcl = 1  --> do derived (non-castable) thing
    // this->castable = 1, includeOtcl = 1  --) castable constructor
    // 	                   includeOtcl = 0  --> castable's newObject args
    int totargs = 0;
    if (!includeOtcl || !castable) {
	int a;
	for (a = 0; a < noOfArgs; a++) {
	    if(totargs)	    of << ", ";
	    arg[a]->type(of);
	    of << " arg" << a;
	    ++totargs;
	}
    }
    if(includeOtcl) {
	if(totargs)	    of << ", ";
	of << "OtclObject *o";
	++totargs;
    }
}

void CdlConstructor::preProcessArgs (ofstream &of, CdlIndent tab)
{
   char varName[32];
   char strName[32];
   int a;
   for (a = 0; a < minNoOfArgs; a++)
   {
     sprintf(varName,"arg%d",a);
     sprintf(strName,"argv[%d]",a);
     arg[a]->initVarFromStr(varName,strName,of,tab,0);
   }
   for (a = minNoOfArgs; a < noOfArgs; a++) {
     sprintf(varName,"arg%d",a);
     if (arg[a]->defaultFlag == ARG_DEFAULT_VAL) {
	 sprintf(strName,"\"%s\"", arg[a]->defaultVal);
	 arg[a]->initVarFromStr(varName,strName,of,tab,1);
     } else {
	 assert(arg[a]->defaultFlag == ARG_DEFAULT_SLOT);
	 sprintf(strName, "part->%s",arg[a]->defaultVal); 
	 arg[a]->initFromVar(varName,strName,of,tab,1);
     }
     of << tab << "if(argc > " << a << ") {" << endl; tab++;
     sprintf(strName,"argv[%d]",a);
     arg[a]->initVarFromStr(varName,strName,of,tab,0,1);
     tab--; of << tab << "}" << endl;
   }
}

void CdlConstructor::passArgs (ofstream &of, char *lastArg)
{
   int a;
   for (a = 0; a < noOfArgs; a++)
   {
      of << "arg" << a << ", ";
   }
   of << lastArg;
}

void CdlConstructor::passArgsUpwards (ofstream &of)
{
   int a;
   for (a = 0; a < noOfArgs; a++)
   {
      of << "arg" << a;
      if (a < (noOfArgs-1))
      {
         of << ", ";
      }
   }
}

void CdlConstructor::postProcessArgs (ofstream &of, CdlIndent tab)
{
   char varName[32];
   char strName[32];
   int a;
   for (a = 0; a < noOfArgs; a++)
   {
      sprintf(varName,"arg%d",a);
      sprintf(strName,"argv[%d]",a);
      arg[a]->cleanVarFromStr(varName,strName,of,tab);
   }
}

int CdlConstructor::isCopyConstructor(char *cls)
{   // If the only constructor specified takes an obref to the 
    // actual type of object, then the code generation just stores
    // the object rather than constructing a new one.  This function 
    // allows the CdlClass to test to see if the constructor is like that.
    if(noOfArgs==1 && strcmp("obref", arg[0]->giveCommand())==0
       && strcmp(cls, ((CdlObrefArg *)arg[0])->giveClass())==0)  {
	return 1;
    } else {
	return 0;
    }
}

int CdlConstructor::hasFromArgs(void)
{    // are any args "-from"? for CdlClass::constructorOK()
    int i;
    for(i=0; i<noOfArgs; ++i) {
	if(arg[i]->defaultFlag == ARG_DEFAULT_SLOT)
	    return 1;
    }
    return 0;
}

CdlClassMethod::CdlClassMethod (char *n) :
   CdlMethod(n)
{
}

CdlClassMethod::~CdlClassMethod ()
{
}

void CdlClassMethod::executeBinding (char *className, char *cname, 
				     ofstream &of, CdlIndent tab, 
				     CdlClass* pclass)
{
    char s[32];
    sprintf(s, "%s::", cname);
    executeBinding_sub(className, cname, of, tab, pclass, s);
}

#ifdef NOTDEF
void CdlClassMethod::executeBinding (char *className, char *cName, 
				     ofstream &of, CdlIndent tab, 
				     CdlClass* pclass)
{
   char varName[32];
   char strName[32];

   of << tab << "if (strcmp(argv[1],\"" << name << "\") == 0)" << endl;
   of << tab << "{" << endl;
   tab++;

   of << tab << "if (argc-2 != " << noOfArgs << ")" << endl;
   of << tab << "{" << endl;
   tab++;
      of << tab << "return Otcl::setTclError(i,ARGS_METHOD_ERR," 
         << "\"" << name << "\",\"" << className << "\");";
   tab--;
   of << tab << "}" << endl;

   int a;
   for (a = 0; a < noOfArgs; a++)
   {
      sprintf(varName,"arg%d",a);
      sprintf(strName,"argv[%d]",a+2);
      arg[a]->initVarFromStr(varName,strName,of,tab,1);
   }  

   if (!voidReturn)
   {
      rtn->declareLocal(of,tab);
   } 
   else
   {
      of << tab << "(void)(";
   }

   of << cName << "::" << name << "(";
   for (a = 0; a < noOfArgs; a++)
   {
      of << "arg" << a;
      if (a < (noOfArgs-1))
      {
         of << ", ";
      }
   }
   of << "));" << endl;
   
   for (a = 0; a < noOfArgs; a++)
   {
      sprintf(varName,"arg%d",a);
      sprintf(strName,"argv[%d]",a+2);
      arg[a]->cleanVarFromStr(varName,strName,of,tab);
   }

   if (!voidReturn)
   {
      rtn->postProcess(of,tab);
   }

   of << tab << "return TCL_OK;" << endl;
   
   tab--;
   of << tab << "}" << endl;
}
#endif
