/*  _ __ ___ _
 * | |\ /  /| |  $Id: Otcl.C,v 1.5 1997/08/11 02:27:02 dpwe Exp dpwe $
 * | | /  / | |  Copyright (C) 1995 IXI Limited.
 * |_|/__/_\|_|  IXI Limited, Cambridge, England.
 *
 * Component   : Otcl.C
 *
 * Author      : Dean Sheehan (deans@x.co.uk)
 *
 * Description : Contains the implementation of the Otcl class that is the
 *               actual extension package for Tcl.
 *
 * License     :
			Object Tcl License & Copyright
			------------------------------

IXI Object Tcl software, both binary and source (hereafter, Software) is copyrighted by IXI Limited (IXI), and ownership remains with IXI. 

IXI grants you (herafter, Licensee) a license to use the Software for academic, research and internal business purposes only, without a fee. Licensee may distribute the binary and source code (if required) to third parties provided that the copyright notice and this statement appears on all copies and that no charge is associated with such copies. 

Licensee may make derivative works. However, if Licensee distributes any derivative work based on or derived from the Software, then Licensee will (1) notify IXI regarding its distribution of the derivative work, and (2) clearly notify users that such derivative work is a modified version and not the original IXI Object Tcl distributed by IXI. IXI strongly recommends that Licensee provide IXI the right to incorporate such modifications into future releases of the Software under these license terms. 

Any Licensee wishing to make commercial use of the Software should contact IXI, to negotiate an appropriate license for such commercial use. Commercial use includes (1) integration of all or part of the source code into a product for sale or license by or on behalf of Licensee to third parties, or (2) distribution of the binary code or source code to third parties that need it to utilize a commercial product sold or licensed by or on behalf of Licensee. 

IXI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY. IXI SHALL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER SUFFERED BY THE USERS OF THIS SOFTWARE. 

Copyright (C) 1995, IXI Limited 

By using or copying this Software, Licensee agrees to abide by the copyright law and all other applicable laws of England and the U.S., including, but not limited to, export control laws, and the terms of this license. IXI shall have the right to terminate this license immediately by written notice upon Licensee's breach of, or non-compliance with, any of its terms. Licensee may be held legally responsible for any copyright infringement that is caused or encouraged by Licensee's failure to abide by the terms of this license. 

Comments and questions are welcome and can be sent to
otcl@x.co.uk 

For more information on copyright and licensing issues, contact: 
Legal Department, IXI Limited, Vision Park, Cambridge CB4 4ZR,
ENGLAND. 

 *
 */

// System Includes
#include <stdio.h>
#include <stdlib.h>

// Tcl Includes
#include <tclInt.h>

#include "Otcl.H"
#include "OtclClass.H"
#include "OtclObjMgr.H"
#include "OtclObject.H"

#include "OtclPart.H"	// for otcl classOf only 

// Class Attribute Definitions
Tcl_Interp *Otcl::tclInterp = NULL;
Otcl *Otcl::otclPtr = NULL;

// To allow post-hoc linking of shared-library Otcl classes (tcl7.5)
extern "C" Otcl* globalOtclPtr;
Otcl* globalOtclPtr = NULL;

Otcl::Otcl (Tcl_Interp *interp, OtclObjMgr *om)
{
   placeCommandsInScope(interp);
   Tcl_InitHashTable(&classes,TCL_STRING_KEYS);

   objMgr = om;

   tclInterp = interp;
   otclPtr = this;
   globalOtclPtr = this;

   // Register all of the available Otcl bound C++ classes
   // with this Otcl extension.
   OtclClassCpp::registerWithOtcl(this);

}

Otcl::~Otcl ()
{
   // Cleanup all the classes in the Hash Table
   Tcl_HashEntry *entry;
   Tcl_HashSearch search;
   OtclClass *otclc;
   for (entry = Tcl_FirstHashEntry(&classes,&search);
        entry != NULL;
        entry = Tcl_NextHashEntry(&search))
   {
      otclc = (OtclClass*)(Tcl_GetHashValue(entry));

      // Not all OtclClass objects registered are owned by this Otcl object
      // as CPP OtclClasses are registered with all Otcl object in this
      // process.
      if (otclc->shouldDelete())
      {
         delete otclc;
      }
   }

   // Cleanup the Hash Table.
   Tcl_DeleteHashTable(&classes);
}

int Otcl::classInterfaceCmd (ClientData cd, Tcl_Interp *interp,
                             int argc, char *argv[])
{
   Otcl *otcl = (Otcl*)cd;
   return otcl->classInterface(interp,argc,argv);
}

int Otcl::classImplementationCmd (ClientData cd, Tcl_Interp *interp,
                                  int argc, char *argv[])
{
   Otcl *otcl = (Otcl*)cd;
   return otcl->classImplementation(interp,argc,argv);
}

int Otcl::instantiateCmd (ClientData cd, Tcl_Interp *interp,
                          int argc, char *argv[])
{
   Otcl *otcl = (Otcl*)cd;
   return otcl->instantiate(interp,argc,argv);
}

int Otcl::discardCmd (ClientData cd, Tcl_Interp *interp,
                      int argc, char *argv[])
{
   Otcl *otcl = (Otcl*)cd;
   return otcl->discard(interp,argc,argv);
}

int Otcl::otclCmd (ClientData cd, Tcl_Interp *interp,
                   int argc, char *argv[])
{
   Otcl *otcl = (Otcl*)cd;
   return otcl->otcl(interp,argc,argv);
}

int Otcl::otclNopCmd (ClientData, Tcl_Interp *interp, int argc, char *argv[])
{
   Tcl_SetResult(interp,Tcl_Merge(argc-1,&argv[1]),TCL_DYNAMIC);
   return TCL_OK;
}

int Otcl::classCmd (ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{
   ARGC_MIN(2)
   {
      return Otcl::setTclError(interp,ARGS_CLASS_CMD_ERR);
   }
   OtclClass *otclc = (OtclClass*)cd;
   return otclc->classMethod(interp,argc,argv);
}

int Otcl::classInterface (Tcl_Interp *interp, int argc, char *argv[])
{
   ARGC_MIN(2)
   {
      return Otcl::setTclError(interp,ARGS_CLASS_INTERFACE_ERR);
   }

   // First argument is the name of the new class (argv[0] is command name)
   int newEntry;
   Tcl_HashEntry *hashEntry = Tcl_CreateHashEntry(&classes,argv[1],&newEntry);
   if (newEntry != 1)
   {
      // Found entry already so must have already interfaced the class!
      return Otcl::setTclError(interp,CLASS_ALREADY_INTERFACED_ERR,argv[1]);
   }

   // Create a new OtclClass object to record the Class interface and later the
   // class internals.
   OtclClassOtcl *otclClass = new OtclClassOtcl(argv[1],this);
   Tcl_SetHashValue(hashEntry,(ClientData)otclClass);

   // Remove the global Otcl commands as we are now about the evaluate
   // the interface commands.
   removeCommandsFromScope(interp);

   int returnCode = otclClass->parseInterface(interp,argc,argv);

   if (returnCode != TCL_OK)
   {
      // Remove the class from the hash table
      Tcl_DeleteHashEntry(hashEntry);
      delete otclClass;
   }

   // Return from evaluating interface so place the global Otcl commands back
   // in scope.
   placeCommandsInScope(interp);

   return returnCode;
}

int Otcl::classImplementation (Tcl_Interp *interp, int argc, char *argv[])
{
   ARGC_MIN(2)
   {
      return Otcl::setTclError(interp,ARGS_CLASS_IMPLEMENTATION_ERR);
   }

   // First argument is the name of the existing class (argv[0] is command nm)
   int newEntry;
   Tcl_HashEntry *hashEntry = Tcl_CreateHashEntry(&classes,argv[1],&newEntry);
   if (newEntry == 1)
   {
      // Couldn't find the entry so couldn't have interfaced the class yet!
      Otcl::setTclResult(interp,CLASS_IMP_BEFORE_INT_ERR,argv[1]);

      // Cleanup the entry that would have been made.
      Tcl_DeleteHashEntry(hashEntry);
      return TCL_ERROR;
   }

   OtclClassOtcl *otclClass = (OtclClassOtcl*)Tcl_GetHashValue(hashEntry);

   // Remove the global Otcl commands as we are now about the evaluate
   // the implementation commands.
   removeCommandsFromScope(interp);

   int returnCode = otclClass->parseImplementation(interp,argc,argv);

   if (returnCode != TCL_OK)
   {
      Tcl_DeleteHashEntry(hashEntry);
      delete otclClass;
   }

   // Return from evaluating implem so place the gloabl Otcl commands back
   // in scope.
   placeCommandsInScope(interp);

   return returnCode;
}

int Otcl::instantiate (Tcl_Interp *interp, int argc, char *argv[])
{
   ARGC_MIN(2)
   {
      return Otcl::setTclError(interp,ARGS_INSTANTIATE_ERR);
   }

   // Look up the class named as the second argument (arg[1]).
   Tcl_HashEntry *hashEntry = Tcl_FindHashEntry(&classes,argv[1]);
   if (hashEntry == NULL)
   {
      Otcl::setTclResult(interp,CLASS_UNDEFINED_ERR,argv[1]);
      return TCL_ERROR;
   }
   OtclClass *otclc = (OtclClass*)Tcl_GetHashValue(hashEntry);

   return otclc->instantiate(interp,argc-2,(argc > 2 ? &argv[2] : NULL),
                             objMgr);
}

int Otcl::discard (Tcl_Interp *interp, int argc, char *argv[])
{
   ARGC_VALUE(2)
   {
      return Otcl::setTclError(interp,ARGS_DISCARD_ERR);
   }

   return discard(interp,argv[1],OTCL_FALSE);
}

int Otcl::discard (Tcl_Interp *interp, char *symRef, int fromCpp)
{
   OtclObject *otclo = objMgr->unManageObject(symRef,interp);
   if (otclo == NULL)
   {
      Otcl::setTclResult(interp,UNKNOWN_OBJECT_ERR,symRef);
      return TCL_ERROR;
   }

   // OtclObject::discard also 'deletes' but it allows us to return
   // values...
   return otclo->discard(interp,fromCpp);
}

int Otcl::otcl (Tcl_Interp *interp, int argc, char *argv[])
{
    ARGC_MIN(2) {
	return Otcl::setTclError(interp,ARGS_OTCL_ERR);
    }

    if (strcmp(argv[1],"clear") == 0) {
	// NOTE!!: This does NOT free all the existing objects, nor 
	// clear their names from the per-object hash tables.  
	// Needs to be brought up to date (recursively call 
	// deleteCmd on all existing objects?? provide a clear
	// method for each class, then call them all??)
	return clear(interp);
    } else if (strcmp(argv[1], "list")==0)  {
	// list the known Otcl classes
	Tcl_HashEntry *entry;
	Tcl_HashSearch search;
	OtclClass *otclc;
	int first = 1;
	for (entry = Tcl_FirstHashEntry(&classes,&search);
	     entry != NULL;
	     entry = Tcl_NextHashEntry(&search))  {
	    otclc = (OtclClass*)(Tcl_GetHashValue(entry));
	    if(first)	first = 0;
	    else		Tcl_AppendResult(interp, " ", NULL);
	    Tcl_AppendResult(interp, otclc->giveName(), NULL);
	}
    } else if (strcmp(argv[1], "classOf")==0)  {
	// return the classname of an argument, if one can be found
	if(argc!=3) {
	    Tcl_AppendResult(interp, "no name passed to 'otcl classOf '", 0);
	    return TCL_ERROR;
	}
	// Use the object manager to find this command's object
	OtclObject *otclo = OtclObjMgrCmd::otclObjOfCmd(interp, argv[2]);
	if(otclo==0)	return TCL_ERROR;	// command set errmsg already
	// Now, just return the class of the object
	OtclPartCpp* otclp = (OtclPartCpp*)otclo->giveOtclPart();
	Tcl_AppendResult(interp, otclp->giveClassName(), NULL);
    } else if (strcmp(argv[1], "_ptr")==0)  {
	// secret function to pass the address of the otcl object to 
	// subsequently-dynamically-loaded objects
	if(argc!=2) {
	    Tcl_AppendResult(interp, "just 'otcl _ptr'", 0);
	    return TCL_ERROR;
	}
	sprintf(interp->result, "%lx", this);
    } else {
	Tcl_AppendResult(interp, "unrecognized subcommand \"", argv[1], 
			 "\" (not clear/list/classOf)", NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

int Otcl::clear (Tcl_Interp *)
{
   // Cleanup all the classes in the Hash Table
   Tcl_HashEntry *entry;
   Tcl_HashSearch search;
   OtclClass *otclc;
   for (entry = Tcl_FirstHashEntry(&classes,&search);
        entry != NULL;
        entry = Tcl_NextHashEntry(&search))
   {
      otclc = (OtclClass*)(Tcl_GetHashValue(entry));

      // Not all OtclClass objects registered are owned by this Otcl object
      // as CPP OtclClasses are registered with all Otcl object in this
      // process.
      if (otclc->shouldDelete())
      {
         delete otclc;
      }
   }

   Tcl_DeleteHashTable(&classes);
   Tcl_InitHashTable(&classes,TCL_STRING_KEYS);

   // Register all of the available Otcl bound C++ classes
   // with this Otcl extension.
   OtclClassCpp::registerWithOtcl(this);

   return TCL_OK;
}

void Otcl::placeCommandsInScope (Tcl_Interp *interp)
{
   Tcl_CreateCommand(interp,CLASS_INTERFACE_CMD,Otcl::classInterfaceCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp,CLASS_IMPLEMENTATION_CMD,
                     Otcl::classImplementationCmd,
                    (ClientData)this,(Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp,INSTANTIATE_CMD,Otcl::instantiateCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp,DISCARD_CMD,Otcl::discardCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp,OTCL_CMD,Otcl::otclCmd,
                     (ClientData)this,(Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand(interp,OTCL_NOP_CMD,Otcl::otclNopCmd,
                     (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);

}

void Otcl::removeCommandsFromScope (Tcl_Interp *interp)
{
   Tcl_DeleteCommand(interp,CLASS_INTERFACE_CMD);
   Tcl_DeleteCommand(interp,CLASS_IMPLEMENTATION_CMD); 
   Tcl_DeleteCommand(interp,INSTANTIATE_CMD); 
   Tcl_DeleteCommand(interp,DISCARD_CMD); 
   Tcl_DeleteCommand(interp,OTCL_CMD);
   Tcl_DeleteCommand(interp,OTCL_NOP_CMD);
}

    // 1995jul11 - major Otcl incompatibility
    // <Class> is now the constructor command (was the class method prefix)
    // <Class>:: is now the class method prefix

// 1997jan15 - <Class>:: interacts badly with Incr Tcl 2.2's scoping operator.
// Use <Class>.. instead, as a vague reminder.

// Standard functions to convert a class name into these two derived 
// command names, so that they are the same when we construct them 
// and when we destroy them.

char *Otcl::giveClassCommandName(char *name)
{
    static char command[64];	// return in a static string - probably safe
    strcpy(command, name);
    //strcat(command, "::");
    // Can't use :: because [incr tcl] v2.2 already traps it - OOPS!
    strcat(command, OTCL_CLASS_CMD_SUFFIX);  // define in Otcl.H
    return command;
}

char *Otcl::giveConstructorName(char *name)
{
    static char command[64];	// return in a static string - probably safe
    strcpy(command, name);
    // strcat(command, "_");
    return command;
}

// #include <iostream.h>     // for debug only

void Otcl::createClassCommands(Tcl_Interp *interp, char *name,OtclClass *otclc)
{   // create the classMethod command and the alternate constructor
    // Has to be done through the Otcl object to get the objMgr field

//	cout << "Otcl::createClassCmds for " << name << endl;

    // These commands must be *destroyed* when the class goes out of scope.
    //  therefore, we need to rebuild the names when we destroy the 
    //  class in OtclClass.C:66 OtclClass::~OtclClass()
    // (SOLVED by Otcl static functions giveClassCommandName and giveCons name)

   Tcl_CreateCommand(interp, Otcl::giveClassCommandName(name), 
		     Otcl::classCmd, (ClientData)otclc, 
		     (Tcl_CmdDeleteProc*)NULL);
   // Create <Object> as a creator command (for Tk-style instancing)
   // ( called from OtclClassOtcl::addClassCommand() )
   // (have to store the otclobjectmanager in the class to be able to 
   //  use the createCmd)
   otclc->otclom = objMgr;
   Tcl_CreateCommand(tclInterp,Otcl::giveConstructorName(name),
		     OtclClass::createCmdCmd,
		     (ClientData)otclc, 
		     (Tcl_CmdDeleteProc*)NULL);

}

#include <iostream.h>	// just for debug messages below

void Otcl::registerOtclClassCpp (OtclClassCpp *otclClassCpp)
{
   int newEntry;
   Tcl_HashEntry *hashEntry = Tcl_CreateHashEntry(&classes,
						  otclClassCpp->giveName(),
                                                  &newEntry);
//cerr << "Otcl::registerOtclClassCpp: " << otclClassCpp->giveName() \
//	<< "@" << (void*) otclClassCpp << endl;
   if (newEntry != 1)
   {
      fprintf(stderr,"Name clash while registering %s C++ class with Object Tcl.\n",otclClassCpp->giveName());
      exit(-1);
   }

   Tcl_SetHashValue(hashEntry,(ClientData)otclClassCpp);

   createClassCommands(tclInterp, otclClassCpp->giveName(), otclClassCpp);

}

// fns in tclVar.c or our OtclTclVar.c
extern "C" Var* TclNewVar(Tcl_Interp* interp);

Var *Otcl::NewVar(Tcl_Interp *tclInterp) {
    // initialize a new Var structure
    return TclNewVar(tclInterp);
}

void Otcl::createProcScope(Tcl_Interp* tclInterp)
{
   Interp *interp = (Interp*)tclInterp;

   // Set up a new call frame
   CallFrame *callFrame = new CallFrame;
   Tcl_InitHashTable(&callFrame->varTable,TCL_STRING_KEYS);
   callFrame->argc = 0;
   callFrame->argv = NULL;
   callFrame->level = (interp->varFramePtr == NULL) ?
                      1 : interp->varFramePtr->level + 1;
   callFrame->callerPtr = interp->framePtr;
   callFrame->callerVarPtr = interp->varFramePtr;
#ifdef ITCL_NAMESPACES
   /* things that need to be set when running under ITCL */
   callFrame->activeNs = interp->activeNs;
   callFrame->activeData = interp->activeData;
   /* callFrame->flags = ITCL_GLOBAL_VAR; */  /* ?? */
   callFrame->flags = 0;
   if(0){
/*    Itcl_ActiveNamespace nsToken = NULL; */
    Itcl_Namespace ns;
    ClientData cdata;
        ns = (Itcl_Namespace)callFrame->activeNs;
        cdata = callFrame->activeData;

        nsToken = Itcl_ActivateNamesp2((Tcl_Interp*)interp, ns, cdata);
   }
#endif /* ITCL */
   interp->framePtr = callFrame;
   interp->varFramePtr = callFrame;
   interp->returnCode = TCL_OK;
}

Itcl_ActiveNamespace Otcl::nsToken = NULL;

void Otcl::destroyProcScope(Tcl_Interp* tclInterp)
{
    Interp* interp = (Interp*)tclInterp;
    CallFrame *topFrame = interp->framePtr;

    // Blow away all local variables
    // Usede to be version of code Taken from TclDeleteVars in tclVar.c 
    // from Tcl7.3 distribution, but substituting Otcl::callTraces 
    // and Otcl::deleteArray.  Since I don't care to support the Otcl_Tcl 
    // parts any more, I'll just call the (current version) of TclDeleteVars 
    // directly.  This takes care of (some) Itcl namespaces stuff as
    // required.

    TclDeleteVars(interp, &topFrame->varTable);

    interp->framePtr = topFrame->callerPtr;
    interp->varFramePtr = topFrame->callerVarPtr;
    
#ifdef ITCL_NAMESPACES
    if(0){
        Itcl_DeactivateNamesp((Tcl_Interp*)tclInterp, nsToken);
    }
#endif /* ITCL_NAMESPACES */

    delete topFrame;
}


OtclClass *Otcl::giveOtclClass (char *name)
{
   Tcl_HashEntry *hashEntry = Tcl_FindHashEntry(&classes,name);
   if (hashEntry != NULL)
   {
      return (OtclClass*)Tcl_GetHashValue(hashEntry);
   }
   return NULL;
}

void *Otcl::obrefToCpp (char *obref, char *cppClassName)
{
   OtclObject *otclObject = objMgr->dereference(obref,tclInterp);
   if (otclObject)
   {
      return otclObject->toCpp(cppClassName);
   }
   else
   {
      return NULL;
   }
}

#if 0
static void
InterpDelete (ClientData clientData,
              Tcl_Interp *interp)
{
    Otcl *obj = (Otcl *) clientData;
    delete obj;
}
#endif

int
Otcl_Init (Tcl_Interp *interp)
{
    Otcl *obj = new Otcl(interp, new OtclObjMgrCmd);
#if 0
    Tcl_CallWhenDeleted (interp, InterpDelete, (ClientData) obj);
#endif
    // for Tcl7.5+, signal the package mechanism
    Tcl_PkgProvide(interp, "Otcl", OTCL_VERSION);

//    cerr << "Otcl_Init executed" << endl;
    return TCL_OK;
}

#ifndef NO_SAFE_TCL
int
Otcl_SafeInit (Tcl_Interp *interp)
{
    return Otcl_Init(interp);
}
#endif /* NO_SAFE_TCL */

#include "OtclPart.H"	// oops - need for calls to extract class attribs

OtclObject *Otcl::createObjectWrapper (OtclPart *part)
{
   OtclObject *otclo = new OtclObject();
   char *name = part->giveClassName();
   int  *pcount = part->giveClassCountPtr();
   char *symbolicRef = otclPtr->objMgr->manageObject(otclo,tclInterp,name,pcount);
   if (symbolicRef == NULL)
   {
      delete otclo;
      return NULL;
   }

   otclo->setSelf(symbolicRef);

   otclo->setPart(part);

   return otclo;
}
